#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from . import TestBase
import amplpy

class TestExceptions(TestBase.TestBase):
    """Test Exceptions."""

    def test_get_option_fail(self):
        ampl = self.ampl
        with self.assertRaises(ValueError):
            ampl.get_option("option 42 42")

    def test_get_value_fail(self):
        ampl = self.ampl
        with self.assertRaises(TypeError):
            ampl.get_value("x.val")

    def test_ampl_couldnotstarted(self):
        env = amplpy.Environment("nodir", "noname")
        with self.assertRaises(RuntimeError):
            ampl = amplpy.AMPL(env)

    def test_export_model_exception(self):
        ampl = self.ampl
        with self.assertRaises(IOError):
            ampl.export_model("/")

    def test_export_data_exception(self):
        ampl = self.ampl
        with self.assertRaises(IOError):
            ampl.export_data("/")

    def test_snapshot_exception(self):
        ampl = self.ampl
        with self.assertRaises(IOError):
            ampl.snapshot("/")

    def test_ampl_cd_exception(self):
        ampl = self.ampl
        with self.assertRaises(RuntimeError):
            ampl.cd("nodir")

    def test_get_variable_exception(self):
        ampl = self.ampl
        ampl.eval("param x;")
        with self.assertRaises(KeyError):
            ampl.get_variable("x")

    def test_get_constraint_exception(self):
        ampl = self.ampl
        ampl.eval("var x;")
        with self.assertRaises(KeyError):
            ampl.get_constraint("x")

    def test_get_objective_exception(self):
        ampl = self.ampl
        ampl.eval("var x;")
        with self.assertRaises(KeyError):
            ampl.get_objective("x")

    def test_get_set_exception(self):
        ampl = self.ampl
        ampl.eval("var x;")
        with self.assertRaises(KeyError):
            ampl.get_set("x")

    def test_get_param_exception(self):
        ampl = self.ampl
        ampl.eval("set x;")
        with self.assertRaises(KeyError):
            ampl.get_parameter("x")

    def test_entitymap_iterator_exception(self):
        ampl = self.ampl
        ampl.eval("param x; param xx; var y;")
        with self.assertRaises(KeyError):
            y = ampl.get_parameters()["y"]

    def test_display_exception(self):
        ampl = self.ampl
        with self.assertRaises(RuntimeError):
            ampl.display("x;")

    def test_ampl_exception_methods(self):
        exc = amplpy.AMPLException("test.mod", 42, 5, "Invalid expression")

        self.assertEqual(exc.get_source_name(), "test.mod")
        self.assertEqual(exc.get_line_number(), 42)
        self.assertEqual(exc.get_offset(), 5)
        self.assertEqual(exc.get_message(), "Invalid expression")

    def test_throw_on_warning(self):
        ampl = self.ampl
        with self.assertRaises(amplpy.AMPLException):
            ampl.eval("c: 3 > 2;")

        ampl.setOption("_throw_on_warnings", False)
        self.assertEqual(ampl.getOption("_throw_on_warnings"), False)
        try:
            ampl.eval("c1: 4 > 2;")
        except amplpy.AMPLException as e:
            self.fail(f"AMPLException was raised unexpectedly: {e}")

    def test_throw_on_warning(self):
        ampl = self.ampl

        try:
            ampl.setOption("times", True)
            ampl.setOption("gentimes", True)

            self.assertEqual(ampl.getOption("times"), True)
            self.assertEqual(ampl.getOption("gentimes"), True)

            ampl.setOption("times", False)
            ampl.setOption("gentimes", False)

            self.assertEqual(ampl.getOption("times"), False)
            self.assertEqual(ampl.getOption("gentimes"), False)
        except Exception as e:
            self.fail(f"Exception was raised unexpectedly: {e}")


    def test_default_errorhandler_error(self):
        ampl = self.ampl
        with self.assertRaises(amplpy.AMPLException) as context:
            ampl.eval("Y Y;")

    def test_default_errorhandler_warning(self):
        ampl = self.ampl
        with self.assertRaises(amplpy.AMPLException) as context:
            ampl.eval("var x := y; var xx := y; var xy := y;")

        self.assertEqual(context.exception.get_message(), 
            "y is not defined\ncontext:  var x :=  >>> y; <<<  var xx := y; var xy := y;")
        
    def test_cons_exceptions(self):
        ampl = self.ampl
        ampl.eval("var x>=0; s.t. cons: x^2 + x >= 42;")
        cons = ampl.get_constraint("cons")
        ampl.reset()
        with self.assertRaises(RuntimeError):
            cons.drop()
        with self.assertRaises(RuntimeError):
            cons.restore()
        with self.assertRaises(TypeError):
            cons.astatus()
        with self.assertRaises(TypeError):
            cons.defvar()
        with self.assertRaises(TypeError):
            cons.dinit()
        with self.assertRaises(TypeError):
            cons.dinit0()
        #with self.assertRaises(AttributeError):
        #    cons.is_logical()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from libcpp cimport bool as bool_c



cdef class Constraint(Entity):
    """
    Bases: :class:`~amplpy.Entity`.

    This class represents an algebraic or logical constraint. In case the
    constraint is scalar, its values can be accessed via functions like
    :func:`~amplpy.Constraint.body` and :func:`~amplpy.Constraint.dual`.
    All the AMPL suffixes for constraints (see
    http://www.ampl.com/NEW/suffbuiltin.html)
    are available through methods of this class with the same name (and methods
    of ConstraintInstance for indexed constraints).

    Note that, since this class represents both algebraic and logical
    constraints, some suffixes might not be available for every entity.

    An TypeError is thrown if one of such methods is called for
    a non-scalar constraint and if a method corresponding to a suffix which is
    not supported by the type of the constraint is called.
    A runtime error is thrown if any property of an entity which has been
    deleted from the underlying interpreter is accessed.

    To gain access to all the values in an entity (for all instances and all
    suffixes for that entities), see
    :func:`~amplpy.Entity.get_values`
    and the :class:`~amplpy.DataFrame` class.
    """
    @staticmethod
    cdef create(AMPL ampl, char* name, campl.AMPL_TUPLE* index, parent):
        entity = Constraint()
        entity._ampl = ampl
        Py_INCREF(entity._ampl)
        entity._name = name
        entity._index = index
        entity.wrap_function = campl.AMPL_CONSTRAINT
        entity._entity = parent
        if entity._entity is not None:
            Py_INCREF(entity._entity)
        return entity

    def __setitem__(self, index, value):
        self.__getitem__(index).set_dual(value)

    def is_logical(self):
        """
        Check if the constraint is a logical constraint. Returns True if
        logical. The available suffixes differ between logical and non logical
        constraints. See http://www.ampl.com/NEW/suffbuiltin.html for a list of
        the available suffixes for algebraic constraints. The suffixes
        available for logical constraints are marked on the method description
        by "Valid only for logical constraints".
        """
        cdef bool_c value
        try:
            PY_AMPL_CALL(campl.AMPL_ConstraintIsLogical(self._ampl._c_ampl, self._name, &value))
            return value
        except AttributeError:
            return False

    def drop(self):
        """
        Drop all instances in this constraint entity, corresponding to the AMPL
        code: `drop constraintname;`.
        """
        PY_AMPL_CALL(campl.AMPL_EntityDrop(self._ampl._c_ampl, self._name))

    def restore(self):
        """
        Restore all instances in this constraint entity, corresponding to the
        AMPL code: `restore constraintname;`.
        """
        PY_AMPL_CALL(campl.AMPL_EntityRestore(self._ampl._c_ampl, self._name))

    def body(self):
        """
        Get the current value of the constraint's body.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_BODY, &value))
        return value

    def astatus(self):
        """
        Get the current AMPL status (dropped, presolved, or substituted out).
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_ASTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def defvar(self):
        """
        Get the index in `_var` of "defined variable" substituted out by the
        constraint.
        """
        cdef int value
        PY_AMPL_CALL(campl.AMPL_InstanceGetIntSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DEFVAR, &value))
        return value

    def dinit(self):
        """
        Get the current initial guess for the constraint's dual variable.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DINIT, &value))
        return value

    def dinit0(self):
        """
        Get the original initial guess for the constraint's dual variable.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DINIT0, &value))
        return value

    def dual(self):
        """
        Get the current value of the constraint's dual variable.

        Note that dual values are often reset by the underlying AMPL
        interpreter by the presolve functionalities triggered by some methods.
        A possible workaround is to set the option `presolve` to `false`
        (see :func:`~amplpy.AMPL.setOption`).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DUAL, &value))
        return value

    def lb(self):
        """
        Get the current value of the constraint's lower bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LB, &value))
        return value

    def ub(self):
        """
        Get the current value of the constraint's upper bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UB, &value))
        return value

    def lbs(self):
        """
        Get the constraint lower bound sent to the solver (reflecting
        adjustment for fixed variables).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LBS, &value))
        return value

    def ubs(self):
        """
        Get the constraint upper bound sent to the solver (reflecting
        adjustment for fixed variables).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UBS, &value))
        return value

    def ldual(self):
        """
        Get the current dual value associated with the lower bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LDUAL, &value))
        return value

    def udual(self):
        """
        Get the current dual value associated with the upper bounds.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UDUAL, &value))
        return value

    def lslack(self):
        """
        Get the slack at lower bound `body - lb`.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LSLACK, &value))
        return value

    def uslack(self):
        """
        Get the slack at upper bound `ub - body`.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_USLACK, &value))
        return value

    def slack(self):
        """
        Constraint slack (the lesser of lslack and uslack).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_SLACK, &value))
        return value

    def sstatus(self):
        """
        Get the solver status (basis status of constraint's slack or artificial
        variable).
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_SSTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def status(self):
        """
        Get the AMPL status if not `in`, otherwise solver status.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_STATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def set_dual(self, dual):
        """
        Set the value of the dual variable associated to this constraint (valid
        only if the constraint is scalar). Equivalent to the AMPL statement:

        `let c := dual;`

        Note that dual values are often reset by the underlying AMPL
        interpreter by the presolve functionalities triggered by some methods.
        A possible workaround is to set the option `presolve` to `False`
        (see :func:`~amplpy.AMPL.setOption`).

        Args:
            dual: The value to be assigned to the dual variable.
        """
        PY_AMPL_CALL(campl.AMPL_ConstraintInstanceSetDual(self._ampl._c_ampl, self._name, self._index, float(dual)))

    def val(self):
        """
        Get the AMPL val suffix. Valid only for logical constraints.
        """
        cdef double value
        if self.is_logical():
            PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_VAL, &value))
            return value
        else:
            return None

    # Aliases
    isLogical = is_logical
    setDual = set_dual



# -*- coding: utf-8 -*-

cdef class Variable(Entity):
    """
    Bases: :class:`~amplpy.Entity`.

    Represents an AMPL decision variable. Note that, in case of a scalar
    variable, all the properties of the variable instance can be accessed
    through methods like :func:`~amplpy.Variable.value`. The methods have the
    same name of the corresponding AMPL suffixes.
    See http://www.ampl.com/NEW/suffbuiltin.html for a list of the available
    suffixes.

    All these methods throw a TypeError if called for a non scalar
    objective and an RuntimeError if called on an entity which has been deleted
    in the underlying intepreter.

    To gain access to all the values in an entity (for all instances and all
    suffixes for that entities), see
    :func:`~amplpy.Entity.get_values`
    and the :class:`~amplpy.DataFrame` class.
    """
    @staticmethod
    cdef create(AMPL ampl, char *name, campl.AMPL_TUPLE* index, parent):
        entity = Variable()
        entity._ampl = ampl
        Py_INCREF(entity._ampl)
        entity._name = name
        entity._index = index
        entity.wrap_function = campl.AMPL_VARIABLE
        entity._entity = parent
        if entity._entity is not None:
            Py_INCREF(entity._entity)
        return entity

    def __setitem__(self, index, value):
        self.__getitem__(index).set_value(value)

    def value(self):
        """
        Get the current value of this variable.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_VALUE, &value))
        return value

    def fix(self, value=None):
        """
        Fix all instances of this variable to a value if provided or to
        their current value otherwise.

        Args:
            value: value to be set.
        """
        if value is None:
            PY_AMPL_CALL(campl.AMPL_VariableInstanceFix(self._ampl._c_ampl, self._name, self._index))
        else:
            PY_AMPL_CALL(campl.AMPL_VariableInstanceFixToValue(self._ampl._c_ampl, self._name, self._index, value))

    def unfix(self):
        """
        Unfix all instances of this variable.
        """
        PY_AMPL_CALL(campl.AMPL_VariableInstanceUnfix(self._ampl._c_ampl, self._name, self._index))

    def set_value(self, value):
        """
        Set the current value of this variable (does not fix it),
        equivalent to the AMPL command `let`.

        Args:
            value: value to be set.
        """
        if isinstance(value, Parameter):
            PY_AMPL_CALL(campl.AMPL_VariableInstanceSetValue(self._ampl._c_ampl, self._name, self._index, value.value()))
        else:
            PY_AMPL_CALL(campl.AMPL_VariableInstanceSetValue(self._ampl._c_ampl, self._name, self._index, float(value)))

    def astatus(self):
        """
        Get the AMPL status (fixed, presolved, or substituted out).
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_ASTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def defeqn(self):
        """
        Get the index in `_con` of "defining constraint" used to substitute
        variable out.
        """
        cdef int value
        PY_AMPL_CALL(campl.AMPL_InstanceGetIntSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DEFEQN, &value))
        return value

    def dual(self):
        """
        Get the dual value on defining constraint of variable substituted out.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_DUAL, &value))
        return value

    def init(self):
        """
        Get the current initial guess.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_INIT, &value))
        return value

    def init0(self):
        """
        Get the original initial guess (set by `:=` or`default` or by a data
        statement).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_INIT0, &value))
        return value

    def lb(self):
        """
        Returns the current lower bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LB, &value))
        return value

    def ub(self):
        """
        Returns the current upper bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UB, &value))
        return value

    def lb0(self):
        """
        Returns the initial lower bounds, from the var declaration.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LB0, &value))
        return value

    def ub0(self):
        """
        Returns the initial upper bound, from the var declaration.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UB0, &value))
        return value

    def lb1(self):
        """
        Returns the weaker lower bound from AMPL's presolve phase.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LB1, &value))
        return value

    def ub1(self):
        """
        Returns the weaker upper bound from AMPL's presolve phase.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UB1, &value))
        return value

    def lb2(self):
        """
        Returns the stronger lower bound from AMPL's presolve phase.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LB2, &value))
        return value

    def ub2(self):
        """
        Returns the stronger upper bound from AMPL's presolve phase.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_UB2, &value))
        return value

    def lrc(self):
        """
        Returns the reduced cost at lower bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LRC, &value))
        return value

    def urc(self):
        """
        Returns the reduced cost at upper bound.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_URC, &value))
        return value

    def lslack(self):
        """
        Return the slack at lower bound (``val - lb``).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_LSLACK, &value))
        return value

    def uslack(self):
        """
        Return the slack at upper bound (``ub - val``).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_USLACK, &value))
        return value

    def rc(self):
        """
        Get the reduced cost (at the nearer bound).
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_RC, &value))
        return value

    def slack(self):
        """
        Returns the bound slack which is the lesser of
        :func:`~amplpy.Variable.lslack` and :func:`~amplpy.Variable.uslack`.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_SLACK, &value))
        return value

    def sstatus(self):
        """
        Solver status (basis status of variable).
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_SSTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def status(self):
        """
        AMPL status if not `in`, otherwise solver status.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_STATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    # Aliases
    setValue = set_value



# -*- coding: utf-8 -*-

cdef class OutputHandler(object):
    """
    Interface to handle the outputs from the calls to any
    function that causes the underlying AMPL interpreter to display a message.
    If an application needs to
    implement customised output handling, it must implement this interface and
    then register an instance with the AMPL API using the
    :func:`~amplpy.AMPL.set_output_handler`
    method.

    Note that errors and warnings are not passed through this interface,
    see :class:`~amplpy.ErrorHandler` for more information.
    """

    def output(self, kind, msg):
        """
        This method is called when AMPL sends some output derived from
        executing a statement.

        Args:
           kind: kind of the output of the AMPL interpreter.

           msg: text of the output by the AMPL interpreter.
        """
        print(msg, end="", flush=True)

cdef void PyOutput(campl.AMPL_OUTPUTKIND kind, const char* msg, void* usrdata) except * with gil:
    PyOutputHandler = <OutputHandler>usrdata
    PyOutputHandler.output(kind, msg.decode('utf-8'))

class Kind(object):
    """
    Represents the type of the output coming from the interpreter.
    """

    WAITING = campl.AMPL_OUTPUT_WAITING
    """
    Output ``prompt2``, returned when incomplete statements are interpreted.
    """

    BREAK = campl.AMPL_OUTPUT_BREAK
    """
    Output ``break``, displayed when an operation is interrupted with SIGINT.
    """

    CD = campl.AMPL_OUTPUT_CD
    """
    Output ``cd``, returned by the ``cd`` function.
    """

    DISPLAY = campl.AMPL_OUTPUT_DISPLAY
    """
    Output ``display``, returned by the ``display`` function.
    """

    EXIT = campl.AMPL_OUTPUT_EXIT
    """
    Output ``exit``, returned as last message from AMPL before exiting the
    interpreter.
    """

    EXPAND = campl.AMPL_OUTPUT_EXPAND
    """
    Output ``expand``, returned by the ``expand`` function.
    """

    LOAD = campl.AMPL_OUTPUT_LOAD
    """
    Output ``load``, returned by the ``load`` function when loading a library.
    """

    OPTION = campl.AMPL_OUTPUT_OPTION
    """
    Output ``option``, returned by the ``option`` function when getting the
    value of an option.
    """

    PRINT = campl.AMPL_OUTPUT_PRINT
    """
    Output ``print``, returned by the ``print`` function when printing values
    from AMPL command line.
    """

    PROMPT = campl.AMPL_OUTPUT_PROMPT  # prompt1 and prompt3
    """
    Output ``prompt1``, normal AMPL prompt.
    """

    SOLUTION = campl.AMPL_OUTPUT_SOLUTION
    """
    Output ``solution``, returned when loading a solution with the command
    ``solution``, contains the solver message.
    """

    SOLVE = campl.AMPL_OUTPUT_SOLVE
    """
    Output ``solve``, returned by the ``solve`` function, contains the solver
    message.
    """

    SHOW = campl.AMPL_OUTPUT_SHOW
    """
    Output ``show``, returned by the ``show`` function.
    """

    XREF = campl.AMPL_OUTPUT_XREF
    """
    Output ``xref``, returned by the ``xref`` function.
    """

    SHELL_OUTPUT = campl.AMPL_OUTPUT_SHELL_OUTPUT
    """
    Output of the AMPL command ``shell``.
    """

    SHELL_MESSAGE = campl.AMPL_OUTPUT_SHELL_MESSAGE
    """
    Messages from the command ``shell``.
    """

    MISC = campl.AMPL_OUTPUT_MISC
    """
    Output ``misc``.
    """

    WRITE_TABLE = campl.AMPL_OUTPUT_WRITE_TABLE
    """
    Messages from the command ``write table``.
    """

    READ_TABLE = campl.AMPL_OUTPUT_READ_TABLE
    """
    Messages from the command ``read table``.
    """

    _READTABLE = campl.AMPL_OUTPUT_READTABLE
    """
    Internal messages from the command ``read table``.
    """

    _WRITETABLE = campl.AMPL_OUTPUT_WRITETABLE
    """
    Internal messages from the command ``write table``.
    """

    BREAKPOINT = campl.AMPL_OUTPUT_BREAKPOINT
    """
    Breakpoint hit.
    """

    CALL = campl.AMPL_OUTPUT_CALL
    """
    Output of a script ``call``.
    """

    CHECK = campl.AMPL_OUTPUT_CHECK
    """
    Output of a ``check`` operation.
    """

    CLOSE = campl.AMPL_OUTPUT_CLOSE
    """
    Output of a ``close`` command for output redirection.
    """

    COMMANDS = campl.AMPL_OUTPUT_COMMANDS
    """
    Output of a ``commands`` call into another file.
    """

    CONTINUE = campl.AMPL_OUTPUT_CONTINUE
    """
    Issued when ``continue`` is encountered.
    """

    DATA = campl.AMPL_OUTPUT_DATA
    """
    Output of a ``data`` command.
    """

    DELETECMD = campl.AMPL_OUTPUT_DELETECMD
    """
    Output of a ``delete`` command.
    """

    DROP = campl.AMPL_OUTPUT_DROP
    """
    Output of a ``drop`` command.
    """

    DROP_OR_RESTORE_ALL = campl.AMPL_OUTPUT_DROP_OR_RESTORE_ALL
    """
    Internal.
    """

    _ELSE = campl.AMPL_OUTPUT_ELSE
    """
    Else block.
    """

    ELSE_CHECK = campl.AMPL_OUTPUT_ELSE_CHECK
    """
    Internal.
    """

    ENDIF = campl.AMPL_OUTPUT_ENDIF
    """
    End of if block.
    """

    ENVIRON = campl.AMPL_OUTPUT_ENVIRON
    """
    Output of a ``environ`` command.
    """

    FIX = campl.AMPL_OUTPUT_FIX
    """
    Output of a ``fix`` command.
    """

    FOR = campl.AMPL_OUTPUT_FOR
    """
    Output of a ``for`` command.
    """

    _IF = campl.AMPL_OUTPUT_IF
    """
    Output of an ``if`` command.
    """

    LET = campl.AMPL_OUTPUT_LET
    """
    Output of a ``let`` command.
    """

    LOOPEND = campl.AMPL_OUTPUT_LOOPEND
    """
    End of loop.
    """

    OBJECTIVE = campl.AMPL_OUTPUT_OBJECTIVE
    """
    Output of an ``objective`` command.
    """

    OPTION_RESET = campl.AMPL_OUTPUT_OPTION_RESET
    """
    Occurs when resetting option values.
    """

    PRINTF = campl.AMPL_OUTPUT_PRINTF
    """
    Output of a ``printf`` command.
    """

    PROBLEM = campl.AMPL_OUTPUT_PROBLEM
    """
    Output of a ``problem`` command.
    """

    PURGE = campl.AMPL_OUTPUT_PURGE
    """
    Output of a ``purge`` command.
    """

    RBRACE = campl.AMPL_OUTPUT_RBRACE
    """
    Occurs when a right brace is encountered.
    """

    READ = campl.AMPL_OUTPUT_READ
    """
    Output of a ``read`` command.
    """

    RELOAD = campl.AMPL_OUTPUT_RELOAD
    """
    Output of a ``reload`` command.
    """

    REMOVE = campl.AMPL_OUTPUT_REMOVE
    """
    Output of a ``remove`` command.
    """

    REPEAT = campl.AMPL_OUTPUT_REPEAT
    """
    Beginning of a repeat loop.
    """

    REPEAT_END = campl.AMPL_OUTPUT_REPEAT_END
    """
    End of a repeat loop.
    """

    RESET = campl.AMPL_OUTPUT_RESET
    """
    Output of a ``reset`` command.
    """

    RESTORE = campl.AMPL_OUTPUT_RESTORE
    """
    Output of a ``restore`` command.
    """

    RUN_ARGS = campl.AMPL_OUTPUT_RUN_ARGS
    """
    Internal.
    """

    SEMICOLON = campl.AMPL_OUTPUT_SEMICOLON
    """
    Internal.
    """

    SSTEP = campl.AMPL_OUTPUT_SSTEP
    """
    Internal.
    """

    THEN = campl.AMPL_OUTPUT_THEN
    """
    Beginning of the ``then`` part of an if statement.
    """

    UNFIX = campl.AMPL_OUTPUT_UNFIX
    """
    Output of an ``unfix`` command.
    """

    UNLOAD = campl.AMPL_OUTPUT_UNLOAD
    """
    Output of an ``unload`` command.
    """

    UPDATE = campl.AMPL_OUTPUT_UPDATE
    """
    Output of an ``update`` command.
    """

    WRITE = campl.AMPL_OUTPUT_WRITE
    """
    Output of a ``write`` command.
    """



cdef class EnvIterator(object):
    cdef campl.AMPL_ENVIRONMENT* env_c
    cdef campl.AMPL_ENVIRONMENTVAR* begin
    cdef campl.AMPL_ENVIRONMENTVAR* end
    cdef campl.AMPL_ENVIRONMENTVAR* iterator

    @staticmethod
    cdef create(campl.AMPL_ENVIRONMENT* env_c):
        envit = EnvIterator()
        envit.env_c = env_c
        envit.begin = NULL
        envit.end = NULL
        envit.iterator = NULL
        campl.AMPL_EnvironmentGetEnvironmentVar(envit.env_c, &envit.begin)
        cdef size_t size
        campl.AMPL_EnvironmentGetSize(envit.env_c, &size)
        envit.end = envit.begin + size
        envit.iterator = envit.begin
        return envit

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterator >= self.end:
            raise StopIteration
        cdef campl.AMPL_ENVIRONMENTVAR* it = self.iterator
        self.iterator += 1
        cdef char* name_c
        cdef char* value_c
        campl.AMPL_EnvironmentVarGetName(it, &name_c)
        campl.AMPL_EnvironmentVarGetValue(it, &value_c)
        name = name_c.decode('utf-8', errors='replace')
        value = value_c.decode('utf-8', errors='replace')
    
        return (name, value)

cdef class EntityMap(object):
    cdef AMPL _ampl
    cdef campl.AMPL_ENTITYTYPE entity_class
    cdef char** begin
    cdef size_t iterator
    cdef size_t _size

    @staticmethod
    cdef create(AMPL ampl, campl.AMPL_ENTITYTYPE entity_class):
        cdef campl.AMPL_ERRORINFO* errorinfo
        entityit = EntityMap()
        entityit._ampl = ampl
        entityit.entity_class = entity_class
        if entity_class == campl.AMPL_VARIABLE:
            errorinfo = campl.AMPL_GetVariables(entityit._ampl._c_ampl, &entityit._size, &entityit.begin)
        elif entity_class == campl.AMPL_CONSTRAINT:
            errorinfo = campl.AMPL_GetConstraints(entityit._ampl._c_ampl, &entityit._size, &entityit.begin)
        elif entity_class == campl.AMPL_OBJECTIVE:
            errorinfo = campl.AMPL_GetObjectives(entityit._ampl._c_ampl, &entityit._size, &entityit.begin)
        elif entity_class == campl.AMPL_SET:
            errorinfo = campl.AMPL_GetSets(entityit._ampl._c_ampl, &entityit._size, &entityit.begin)
        elif entity_class == campl.AMPL_PARAMETER:
            errorinfo = campl.AMPL_GetParameters(entityit._ampl._c_ampl, &entityit._size, &entityit.begin)
        else:
            raise ValueError(f"Unknown entity class.")

        if errorinfo:
            for i in range(entityit._size):
                campl.AMPL_StringFree(&entityit.begin[i])
            free(entityit.begin)
            PY_AMPL_CALL(errorinfo)

        entityit.iterator = 0
        return entityit

    def __dealloc__(self):
        if self.iterator < self._size:
            for i in range(self.iterator, self._size):
                campl.AMPL_StringFree(&self.begin[i])
        if self.begin != NULL:
            free(self.begin)
            self.begin = NULL

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterator >= self._size:
            raise StopIteration
        tuple = (self.begin[self.iterator].decode('utf-8'), create_entity(self.entity_class, self._ampl, self.begin[self.iterator], NULL, None))
        self.iterator += 1
        return tuple

    def __getitem__(self, key):
        assert isinstance(key, str)
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef campl.AMPL_ENTITYTYPE entitytype
        cdef char* name_c = strdup(key.encode('utf-8'))
        errorinfo = campl.AMPL_EntityGetType(self._ampl._c_ampl, name_c, &entitytype)
        if errorinfo:
            free(name_c)
            PY_AMPL_CALL(errorinfo)
        if entitytype != self.entity_class:
            free(name_c)
            raiseKeyError(self.entity_class, key)
        return create_entity(self.entity_class, self._ampl, name_c, NULL, None)

    def size(self):
        return int(self._size)

    def __len__(self):
        return self.size()

cdef class InstanceIterator(object):
    cdef AMPL _ampl
    cdef char* _name
    cdef campl.AMPL_ENTITYTYPE entity_class
    cdef campl.AMPL_TUPLE** begin
    cdef size_t iterator
    cdef size_t _size
    cdef object _entity

    @staticmethod
    cdef create(AMPL ampl, char* name, campl.AMPL_ENTITYTYPE entity_class, parent):
        cdef campl.AMPL_ERRORINFO* errorinfo
        instanceit = InstanceIterator()
        cdef size_t arity
        instanceit._ampl = ampl
        instanceit._name = name
        instanceit.entity_class = entity_class
        instanceit._entity = parent
        PY_AMPL_CALL(campl.AMPL_EntityGetIndexarity(instanceit._ampl._c_ampl, instanceit._name, &arity))
        if arity == 0:
            instanceit._size = 1
            instanceit.begin = NULL
            instanceit.iterator = 0
            return instanceit
        errorinfo = campl.AMPL_EntityGetTuples(instanceit._ampl._c_ampl, instanceit._name, &instanceit.begin, &instanceit._size)
        if errorinfo:
            for i in range(instanceit._size):
                campl.AMPL_TupleFree(&instanceit.begin[i])
            free(instanceit.begin)
            PY_AMPL_CALL(errorinfo)
        if instanceit._size == 0:
            instanceit.iterator = 0
        else:
            instanceit.iterator = 0
        return instanceit

    def __dealloc__(self):
        if self.iterator+1 < self._size:
            for i in range(self.iterator+1, self._size):
                campl.AMPL_TupleFree(&self.begin[i])
        if self.begin != NULL:
            free(self.begin)
            self.begin = NULL

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterator >= self._size:
            raise StopIteration
        self.iterator += 1
        if self.begin == NULL:
            return (None, self._entity)
        else:
            return (to_py_tuple(self.begin[self.iterator-1]), create_entity(self.entity_class, self._ampl, self._name, self.begin[self.iterator-1], self._entity))

    def __getitem__(self, key):
        assert isinstance(key, str)
        key = tuple(key)
        cdef campl.AMPL_TUPLE* tuple_c = to_c_tuple(key)
        return create_entity(self.entity_class, self._ampl, self._name, tuple_c, self._entity)

    def size(self):
        return int(self._size)

    def __len__(self):
        return self.size()


cdef class MemberRangeIterator(object):
    """Iterator for set members."""
    cdef campl.AMPL* _c_ampl
    cdef char* _name
    cdef campl.AMPL_TUPLE* _index
    cdef campl.AMPL_TUPLE** begin
    cdef size_t iterator
    cdef size_t _size
    cdef object _entity

    @staticmethod
    cdef create(campl.AMPL* ampl, char* name, campl.AMPL_TUPLE* index, parent):
        cdef campl.AMPL_ERRORINFO* errorinfo
        instanceit = MemberRangeIterator()
        instanceit._c_ampl = ampl
        instanceit._name = name
        instanceit._index = index
        instanceit._entity = parent
        if instanceit._entity is not None:
            Py_INCREF(instanceit._entity)
        errorinfo = campl.AMPL_SetInstanceGetValues(instanceit._c_ampl, instanceit._name, instanceit._index, &instanceit.begin, &instanceit._size)
        if errorinfo:
            for i in range(instanceit._size):
                campl.AMPL_TupleFree(&instanceit.begin[i])
            free(instanceit.begin)
            PY_AMPL_CALL(errorinfo)
        instanceit.iterator = 0
        return instanceit

    def __dealloc__(self):
        if self._entity is not None:
            Py_DECREF(self._entity)
        for i in range(self._size):
            campl.AMPL_TupleFree(&self.begin[i])
        if self.begin != NULL:
            free(self.begin)
            self.begin = NULL

    def size(self):
        return int(self._size)

    def __len__(self):
        return self.size()

    def __iter__(self):
        return self

    def __next__(self):
        if self.iterator >= self._size:
            raise StopIteration
        cdef campl.AMPL_TUPLE* it = self.begin[self.iterator]
        cdef size_t size
        cdef campl.AMPL_VARIANT* variant
        campl.AMPL_TupleGetSize(it, &size)
        if size == 1:
            campl.AMPL_TupleGetVariant(it, 0, &variant)
            py_variant = to_py_variant(variant)
            self.iterator += 1
            return py_variant
        else:
            py_tuple = to_py_tuple(it)
            self.iterator += 1
            return py_tuple

cdef class ColIterator(object):
    cdef campl.AMPL_DATAFRAME* _df
    cdef size_t _index
    cdef size_t _rowsize
    cdef size_t _rowit

    @staticmethod
    cdef create(campl.AMPL_DATAFRAME* df, size_t index):
        colit = ColIterator()
        colit._df = df
        colit._index = index
        colit._rowit = 0
        PY_AMPL_CALL(campl.AMPL_DataFrameGetNumRows(colit._df, &colit._rowsize))
        return colit

    def __iter__(self):
        return self

    def __next__(self):
        cdef campl.AMPL_VARIANT* v
        if self._rowit >= self._rowsize:
            raise StopIteration
        PY_AMPL_CALL(campl.AMPL_DataFrameElement(self._df, self._rowit, self._index, &v))
        self._rowit += 1
        return to_py_variant(v)

cdef class RowIterator(object):
    cdef campl.AMPL_DATAFRAME* _df
    cdef size_t _index
    cdef size_t _columnsize
    cdef size_t _columnit

    @staticmethod
    cdef create(campl.AMPL_DATAFRAME* df, size_t index):
        rowit = RowIterator()
        rowit._df = df
        rowit._index = index
        rowit._columnit = 0
        PY_AMPL_CALL(campl.AMPL_DataFrameGetNumCols(rowit._df, &rowit._columnsize))
        return rowit

    def __iter__(self):
        return self

    def __next__(self):
        cdef campl.AMPL_VARIANT* v
        if self._columnit >= self._columnsize:
            raise StopIteration
        PY_AMPL_CALL(campl.AMPL_DataFrameElement(self._df, self._index, self._columnit, &v))
        self._columnit += 1
        return to_py_variant(v)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from . import TestBase
import amplpy


class TestOutputHandler(TestBase.TestBase):
    """Test OutputHandlers."""

    def test_output_kind(self):
        ampl = self.ampl
        class LastOutput(amplpy.OutputHandler):
            def __init__(self):
                self.lastkind = None
                super(LastOutput, self).__init__()

            def output(self, kind, msg):
                self.lastkind = kind

            def getLastKind(self):
                return self.lastkind

        output_handler = LastOutput()
        ampl.set_output_handler(output_handler)
        ampl.eval("print '1';")
        self.assertEqual(output_handler.getLastKind(), amplpy.Kind.PRINT)
        ampl.eval("printf '1';")
        self.assertEqual(output_handler.getLastKind(), amplpy.Kind.PRINTF)
        ampl.eval("var x; display x;")
        self.assertEqual(output_handler.getLastKind(), amplpy.Kind.DISPLAY)
        ampl.eval("show x;")
        self.assertEqual(output_handler.getLastKind(), amplpy.Kind.SHOW)
        ampl.eval("solve;")
        self.assertEqual(output_handler.getLastKind(), amplpy.Kind.SOLVE)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import os


cdef class Environment(object):
    """
    This class provides access to the environment variables and provides
    facilities to specify where to load the underlying AMPL interpreter.
    """
    cdef campl.AMPL_ENVIRONMENT* _c_env

    def __cinit__(self, binary_directory=None, binary_name=None):
        """
        Constructor with ability to select the location of the AMPL binary.
        Note that if binaryDirectory is set, the automatic lookup for an AMPL
        executable will not be executed.

        Args:
            binary_directory: The directory in which look for the AMPL executable.
            binary_name: The name of the AMPL executable.
        """
        if binary_directory is None:
            binary_directory = ""
        if binary_name is None:
            binary_name = ""
        campl.AMPL_EnvironmentCreate(&self._c_env, binary_directory.encode('utf-8'), binary_name.encode('utf-8'))
        if os.name == "nt":
            # Workaround for Windows issue with environment variables
            for option_name in ["ampl_libpath", "solver"]:
                option_value = os.environ.get(option_name, "")
                if option_value:
                    campl.AMPL_EnvironmentAddEnvironmentVariable(self._c_env, option_name.encode('utf-8'), option_value.encode('utf-8'))

    cdef campl.AMPL_ENVIRONMENT* get_c_pointer(self):
        return self._c_env

    def __dealloc__(self):
        campl.AMPL_EnvironmentFree(&self._c_env)

    def __str__(self):
        return self.to_string()

    def __iter__(self):
        return EnvIterator.create(self._c_env)

    def __setitem__(self, name, value):
        """
        Add an environment variable to the environment, or change its value if
        already defined.

        Args:
            name: Name of the environment variable.

            value: Value to be assigned.
        """
        campl.AMPL_EnvironmentAddEnvironmentVariable(self._c_env, name.encode('utf-8'), value.encode('utf-8'))

    def __getitem__(self, name):
        """
        Searches the current object for an environment variable called name and
        returns an iterator to it if found, otherwise it returns `None`.
        """
        cdef campl.AMPL_ENVIRONMENTVAR* iterator
        cdef char* value_c
        if not campl.AMPL_EnvironmentFindEnvironmentVar(self._c_env, name.encode('utf-8'), &iterator):
            campl.AMPL_EnvironmentVarGetValue(iterator, &value_c)
            value = str(value_c.decode('utf-8'))
            return value
        else:
            return None

    def to_string(self):
        cdef char* to_string_c
        campl.AMPL_EnvironmentToString(self._c_env, &to_string_c)
        to_string = str(to_string_c.decode('utf-8'))
        return to_string

    def set_bin_dir(self, binary_directory):
        """
        Set the location where AMPL API will search for the AMPL executable.

        Args:
            binary_directory: The directory in which look for the AMPL executable.
        """
        campl.AMPL_EnvironmentSetBinaryDirectory(self._c_env, binary_directory.encode('utf-8'))

    def get_bin_dir(self):
        """
        Get the location where AMPL API will search for the AMPL executable.
        """
        cdef char* bin_dir_c
        campl.AMPL_EnvironmentGetBinaryDirectory(self._c_env, &bin_dir_c)
        bin_dir = str(bin_dir_c.decode('utf-8'))
        return bin_dir

    def set_bin_name(self, binary_name):
        """
        Set the name of the AMPL executable.

        Args:
            binary_name: The name of the AMPL executable.
        """
        campl.AMPL_EnvironmentSetBinaryName(self._c_env, binary_name.encode('utf-8'))

    def get_bin_name(self):
        """
        Get the name of the AMPL executable.
        """
        cdef char* bin_name_c
        campl.AMPL_EnvironmentGetBinaryName(self._c_env, &bin_name_c)
        bin_name = str(bin_name_c.decode('utf-8'))
        return bin_name
    
    # Aliases
    toString = to_string
    getBinDir = get_bin_dir
    getBinName = get_bin_name
    setBinDir = set_bin_dir
    setBinName = set_bin_name



# -*- coding: utf-8 -*-
from libc.stdlib cimport malloc, free
from libc.string cimport strdup

from cpython.float cimport PyFloat_AsDouble
from cpython.long cimport PyLong_AsLong, PyLong_Check
from cpython.exc cimport PyErr_Occurred, PyErr_Clear
from libc.stdint cimport int64_t

from cpython.pycapsule cimport PyCapsule_IsValid

from numbers import Real

import nanoarrow as na
from nanoarrow import c_array, c_schema, c_array_stream


try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import pyarrow as pa
except ImportError:
    pa = None
try :
    import polars as pl
except ImportError:
    pl = None


cdef class DataFrameArrow:
    cdef campl.AMPL_DATAFRAMEARROW* _c_df
    cdef object _capsule_schema
    cdef object _capsule_array

    def __cinit__(self):
        self._c_df = NULL

    def __dealloc__(self):
        if self._c_df != NULL:
            # Uncomment when Free is implemented
            # campl.AMPL_DataFrameArrowFree(&self._c_df)
            self._c_df = NULL


    cdef void init_from_arrow(self,
                              campl.ArrowSchema* schema_ptr,
                              campl.ArrowArray* array_ptr,
                              int64_t nindices):
        #if array_ptr.release == NULL:
        #    raise ValueError("ArrowArray has been released — invalid memory.")
        cdef campl.AMPL_ERRORINFO* errorinfo
        errorinfo = campl.AMPL_DataFrameArrowCreate(&self._c_df,
                                                    schema_ptr,
                                                    array_ptr,
                                                    nindices)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)

    cdef campl.AMPL_DATAFRAMEARROW* get_ptr(self):
        return self._c_df

    @classmethod
    def from_polars(cls, df):
        """
        Create a :class:`~amplpy.DataFrameArrow` from a Polars DataFrame.
        """
        assert pl is not None and isinstance(df, pl.DataFrame)
        stream = na.c_array_stream(df.to_arrow())

        c_schema_capsule = stream.get_schema().__arrow_c_schema__()
        _, c_array_capsule = stream.get_next().__arrow_c_array__()

        cdef const campl.ArrowSchema* schema_ptr = <const campl.ArrowSchema*>PyCapsule_GetPointer(c_schema_capsule, "arrow_schema")
        cdef const campl.ArrowArray* array_ptr = <const campl.ArrowArray*>PyCapsule_GetPointer(c_array_capsule, "arrow_array")

        cdef DataFrameArrow obj = DataFrameArrow()

        obj.init_from_arrow(schema_ptr, array_ptr, 0)
        return obj

    @classmethod
    def from_pandas(cls, df, index_names=None, indexarity=None):
        """
        Create a :class:`~amplpy.DataFrameArrow` from a pandas DataFrame.

        Args:
            df: Pandas DataFrame or Series to load.
            index_names: Optional list of index column names.
            indexarity: Optional number of index columns.
        """
        assert None not in (pd, pa, na) and isinstance(df, (pd.DataFrame, pd.Series))

        if isinstance(df, pd.Series):
            df = pd.DataFrame(df)

        if len(df) == 0:
            raise ValueError("Cannot create DataFrameArrow from empty DataFrame")

        if isinstance(df.index[0], tuple):
            df.index = pd.MultiIndex.from_tuples(df.index.tolist())
        elif df.index.nlevels == 1 and not isinstance(df.index, pd.MultiIndex):
            df.index = pd.MultiIndex.from_tuples([(idx,) for idx in df.index])

        if indexarity is not None and indexarity == df.index.nlevels + 1:
            df = df.stack()
            if isinstance(df, pd.Series):
                df = pd.DataFrame(df)

        if index_names is not None:
            assert len(index_names) == df.index.nlevels
            df.index.names = index_names

        df_reset = df.reset_index()

        pa_df = pa.Table.from_pandas(df_reset, preserve_index=False)
        #print(pa_df)
        array_stream = na.c_array_stream(pa_df)
        c_schema_capsule, c_array_capsule = array_stream.get_next().__arrow_c_array__()

        cdef campl.ArrowSchema* arrow_schema_ptr = <campl.ArrowSchema*>PyCapsule_GetPointer(c_schema_capsule, "arrow_schema")
        cdef campl.ArrowArray* arrow_array_ptr = <campl.ArrowArray*>PyCapsule_GetPointer(c_array_capsule, "arrow_array")

        # VALIDATE capsule contents
        #if not PyCapsule_IsValid(c_schema_capsule, "arrow_schema"):
        #    raise RuntimeError("Invalid arrow_schema capsule")
        #if not PyCapsule_IsValid(c_array_capsule, "arrow_array"):
        #    raise RuntimeError("Invalid arrow_array capsule")

        cdef DataFrameArrow obj = DataFrameArrow()
        obj._capsule_schema = c_schema_capsule
        obj._capsule_array = c_array_capsule

        cdef int64_t nindices = df.index.nlevels if indexarity is None else indexarity
        obj.init_from_arrow(arrow_schema_ptr, arrow_array_ptr, nindices)
        return obj



# -*- coding: utf-8 -*-
import logging

try:
    from .tools import _SUPPORT_MESSAGE
except Exception:
    _SUPPORT_MESSAGE = ""

logger = logging.getLogger("amplpy")
logger.setLevel(logging.WARNING)

if not logger.hasHandlers():
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(levelname)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False

def display_error_message(exception, error=True):
    logger.setLevel(logging.WARNING)
    msg = "\t" + str(exception).replace(_SUPPORT_MESSAGE, "").replace("\n", "\n\t")
    if error:
        logger.error(f"\n{msg}")
    else:
        logger.warning(f"\n{msg}")


cdef class ErrorHandler:
    """
    A basic interface for AMPL error handlers. If an application needs to
    implement customized error handling, it must implement this interface and
    then register an instance with the AMPL API using the
    :func:`~amplpy.AMPL.set_error_handler` method.
    The underlying AMPL interpreter will then report all errors and warnings
    through this interface as :class:`~amplpy.AMPLException` objects.
    """

    def error(self, exception):
        """
        Receives notification of an error.
        """
        display_error_message(exception)
        raise exception

    def warning(self, exception):
        """
        Receives notification of a warning.
        """
        display_error_message(exception, error=False)
        raise exception

cdef void PyError(bool_c isWarning, const char* filename, int row, int offset, const char* message, void* errorHandler) except * with gil:
    handler = <ErrorHandler>errorHandler
    if isWarning:
        exception = AMPLException(filename.decode('utf-8'), row, offset, message.decode('utf-8'))
        handler.warning(exception)
    else:
        exception = AMPLException(filename.decode('utf-8'), row, offset, f"{message.decode('utf-8')}{_SUPPORT_MESSAGE}")
        handler.error(exception)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from . import TestBase


class TestProperties(TestBase.TestBase):
    """Test Properties."""

    def test_var_prop(self):
        ampl = self.ampl
        ampl.eval("var x;")
        ampl.var["x"] = 10

        self.assertEqual(ampl.var["x"].value(), 10)
        self.assertEqual(ampl.get_variable("x").value(), 10)
        self.assertEqual(len(list(ampl.var)), 1)
        self.assertEqual(len(list(ampl.get_variables())), 1)

    def test_con_prop(self):
        ampl = self.ampl
        ampl.eval("var x; s.t. c: x = 10;")
        self.assertEqual(ampl.con["c"].body(), 10)
        self.assertEqual(ampl.get_constraint("c").body(), 10)
        self.assertEqual(len(list(ampl.con)), 1)
        self.assertEqual(len(list(ampl.get_constraints())), 1)

    def test_con_dual(self):
        ampl = self.ampl
        ampl.eval("var x; var y; s.t. c: x + y >= 42;")
        ampl.con["c"].set_dual(1)
        self.assertEqual(ampl.con["c"].dual(), 1)
        ampl.con["c"] = 42
        self.assertEqual(ampl.con["c"].dual(), 42)

    def test_obj_prop(self):
        ampl = self.ampl
        ampl.eval("var x; maximize obj: x;")
        self.assertEqual(ampl.obj["obj"].name(), "obj")
        self.assertEqual(ampl.get_objective("obj").name(), "obj")
        self.assertEqual(len(list(ampl.obj)), 1)
        self.assertEqual(len(list(ampl.get_objectives())), 1)

    def test_set_prop(self):
        ampl = self.ampl
        ampl.eval("set s;")
        ampl.set["s"] = [1, 2, 3]
        self.assertEqual(ampl.set["s"].name(), "s")
        self.assertEqual(ampl.get_set("s").name(), "s")
        self.assertEqual(ampl.set["s"].size(), 3)
        self.assertEqual(len(list(ampl.set)), 1)
        self.assertEqual(len(list(ampl.get_sets())), 1)

    def test_param_prop(self):
        ampl = self.ampl
        ampl.eval("param p1; param p2{1..2};")
        self.assertEqual(ampl.param["p1"].name(), "p1")
        self.assertEqual(ampl.get_parameter("p1").name(), "p1")
        ampl.param["p1"] = 3
        self.assertEqual(ampl.param["p1"].value(), 3)
        ampl.param["p2"] = {1: 10, 2: 20}
        self.assertEqual(ampl.param["p2"][1], 10)
        self.assertEqual(ampl.param["p2"][2], 20)
        self.assertEqual(len(list(ampl.param)), 2)
        self.assertEqual(len(list(ampl.get_parameters())), 2)

    def test_option_prop(self):
        ampl = self.ampl
        ampl.option["solver"] = "gurobi"
        self.assertEqual(ampl.option["solver"], "gurobi")
        self.assertEqual(ampl.get_option("solver"), "gurobi")

    def test_option_prop_setitem_dict(self):
        ampl = self.ampl
        ampl.option["gurobi_options"] = {
            "outlev": True,
            "alg:method": 2,
            "cvt:mip:eps": 1e-3,
        }
        self.assertEqual(ampl.get_option("gurobi_options"), "outlev=1 alg:method=2 cvt:mip:eps=0.001")

    def test_option_prop_dict(self):
        ampl = self.ampl
        ampl.option = {
            "solver": "gurobi",
            "presolve": False,
            "pl_linearize": False,
            "gurobi_options": {"outlev": True, "timelim": 1},
        }
        self.assertEqual(ampl.get_option("solver"), "gurobi")
        self.assertEqual(ampl.get_option("presolve"), False)
        self.assertEqual(ampl.get_option("pl_linearize"), False)
        self.assertEqual(ampl.get_option("gurobi_options"), "outlev=1 timelim=1")

    def test_solve_result(self):
        ampl = self.ampl
        self.assertEqual(ampl.solve_result, "?")
        self.assertEqual(ampl.solve_result_num, -1)
        ampl.solve()
        self.assertEqual(ampl.solve_result, "solved")
        self.assertEqual(ampl.solve_result_num, 0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import sys
import os
import platform
import ctypes


if platform.system().startswith(("Windows", "MSYS", "CYGWIN", "MINGW")):
    libbase = os.path.join(os.path.dirname(__file__), "amplpython", "cppinterface", "lib")
    lib64 = os.path.join(libbase, "amd64")
    from glob import glob

    try:
        dllfile = glob(lib64 + "/*.dll")[0]
        ctypes.CDLL(dllfile)
    except Exception as exp:
        raise exp

from amplpy.ampl import OutputHandler, Kind
from amplpy.ampl import ErrorHandler
from amplpy.ampl import AMPLException, PresolveException, InfeasibilityException
from amplpy.ampl import EntityMap
from amplpy.ampl import Objective
from amplpy.ampl import Variable
from amplpy.ampl import Constraint
from amplpy.ampl import Set
from amplpy.ampl import Parameter
from amplpy.ampl import Entity
from amplpy.ampl import DataFrame, Row, Column
from amplpy.ampl import DataFrameArrow
from .utils import add_to_path, multidict, register_magics
from amplpy.ampl import Environment
from amplpy.ampl import AMPL
from amplpy.ampl import logger

_parent_dir = os.path.abspath(os.path.dirname(__file__))
_vendor_dir = os.path.join(_parent_dir, "vendor")
if _vendor_dir not in sys.path:
    sys.path.append(_vendor_dir)

try:
    from ampltools import register_magics
except Exception:
    pass
try:
    from ampltools import ampl_notebook
except Exception:
    pass
try:
    from ampltools import add_to_path
except Exception:
    pass
try:
    from . import modules

    modules.preload()
except Exception:
    pass

__version__ = "0.16.0b0"

def _list_aliases():
    from inflection import camelize

    classes = [
        OutputHandler,
        ErrorHandler,
        AMPLException,
        EntityMap,
        Entity,
        Objective,
        Variable,
        Constraint,
        Set,
        Parameter,
        Row,
        Column,
        DataFrame,
        DataFrameArrow,
        Environment,
        AMPL,
    ]
    for cls in classes:
        print(cls)
        for method in list(dir(cls)):
            if method.startswith("__"):
                continue
            cammel_method = camelize(method, False)
            if cammel_method != method:
                print(f"\t{cammel_method} = {method}")
                # setattr(cls, cammel_method, getattr(cls, method))



# pip install autopxd2
# autopxd -I amplpy/amplpython/cppinterface/include/ amplpy/amplpython/cppinterface/include/ampl/ampl_c.h

# add except * to the end of ctypedef void (*ErrorHandlerCbPtr)(bool isWarning, const char* filename, int row, int offset, const char* message, void* errorHandler) except *
# add except * to the end of ctypedef void (*AMPL_OutputHandlerCb)(AMPL_OUTPUTKIND, const char*, void*) except *
# add except * to the end of AMPL_Eval(AMPL* ampl, const char* statement) except *

from libcpp cimport bool # add this line
from libc.stdint cimport int64_t

cdef extern from "ampl/ampl_c.h":

    ctypedef enum AMPL_TYPE:
        AMPL_EMPTY
        AMPL_NUMERIC
        AMPL_STRING

    ctypedef enum AMPL_STRINGSUFFIX:
        AMPL_ASTATUS
        AMPL_SSTATUS
        AMPL_STATUS
        AMPL_MESSAGE
        AMPL_RESULT
        AMPL_SENSE

    ctypedef enum AMPL_NUMERICSUFFIX:
        AMPL_VALUE
        AMPL_DEFEQN
        AMPL_DUAL
        AMPL_INIT
        AMPL_INIT0
        AMPL_LB
        AMPL_UB
        AMPL_LB0
        AMPL_UB0
        AMPL_LB1
        AMPL_UB1
        AMPL_LB2
        AMPL_UB2
        AMPL_LRC
        AMPL_URC
        AMPL_LSLACK
        AMPL_USLACK
        AMPL_RC
        AMPL_SLACK
        AMPL_BODY
        AMPL_DEFVAR
        AMPL_DINIT
        AMPL_DINIT0
        AMPL_LBS
        AMPL_UBS
        AMPL_LDUAL
        AMPL_UDUAL
        AMPL_VAL
        AMPL_EXITCODE

    const char* NUMERICSUFFIXES[29]

    const char* STRINGSUFFIXES[6]

    #ctypedef AMPL_Args AMPL_ARGS
    ctypedef struct AMPL_ARGS:
        pass

    void retainArgs(AMPL_ARGS* args)

    void releaseArgs(AMPL_ARGS* args)

    int AMPL_ArgsCreateNumeric(AMPL_ARGS** args, const double* values)

    int AMPL_ArgsCreateString(AMPL_ARGS** args, const char* const* values)

    int AMPL_ArgsCopy(AMPL_ARGS** args, AMPL_ARGS* copy)

    int AMPL_ArgsDestroy(AMPL_ARGS** args)

    int AMPL_ArgsGetType(AMPL_ARGS* args, AMPL_TYPE* type)

    int AMPL_ArgsGetDblValues(AMPL_ARGS* args, const double** values)

    int AMPL_ArgsGetStrValues(AMPL_ARGS* args, const char* const** values)

    ctypedef enum AMPL_ERRORCODE:
        AMPL_EXCEPTION
        AMPL_LICENSE_EXCEPTION
        AMPL_FILE_IO_EXCEPTION
        AMPL_UNSUPPORTED_OPERATION_EXCEPTION
        AMPL_INVALID_SUBSCRIPT_EXCEPTION
        AMPL_SYNTAX_ERROR_EXCEPTION
        AMPL_NO_DATA_EXCEPTION
        AMPL_LOGIC_ERROR
        AMPL_RUNTIME_ERROR
        AMPL_INVALID_ARGUMENT
        AMPL_OUT_OF_RANGE
        AMPL_STD_EXCEPTION
        AMPL_PRESOLVE_EXCEPTION
        AMPL_INFEASIBILITY_EXCEPTION

    #ctypedef AMPL_ErrorInfo AMPL_ERRORINFO
    ctypedef struct AMPL_ERRORINFO:
        pass

    AMPL_ERRORCODE AMPL_ErrorInfoGetError(AMPL_ERRORINFO* error)

    char* AMPL_ErrorInfoGetMessage(AMPL_ERRORINFO* error)

    int AMPL_ErrorInfoGetLine(AMPL_ERRORINFO* error)

    int AMPL_ErrorInfoGetOffset(AMPL_ERRORINFO* error)

    char* AMPL_ErrorInfoGetSource(AMPL_ERRORINFO* error)

    ctypedef void (*ErrorHandlerCbPtr)(bool isWarning, const char* filename, int row, int offset, const char* message, void* errorHandler) except *

    #ctypedef AMPL_Variant AMPL_VARIANT
    ctypedef struct AMPL_VARIANT:
        pass

    void retainVariant(AMPL_VARIANT* v)

    void releaseVariant(AMPL_VARIANT* v)

    int AMPL_VariantCreateEmpty(AMPL_VARIANT** v)

    int AMPL_VariantCreateNumeric(AMPL_VARIANT** v, double value)

    int AMPL_VariantCreateString(AMPL_VARIANT** v, const char* cstr)

    int AMPL_VariantCopy(AMPL_VARIANT** v, AMPL_VARIANT* copy)

    int AMPL_VariantFree(AMPL_VARIANT** v)

    int AMPL_VariantCompare(AMPL_VARIANT* v1, AMPL_VARIANT* v2)

    int AMPL_VariantGetNumericValue(AMPL_VARIANT* v, double* value)

    int AMPL_VariantGetStringValue(AMPL_VARIANT* v, char** value)

    int AMPL_VariantGetType(AMPL_VARIANT* v, AMPL_TYPE* type)

    int AMPL_VariantFormat(AMPL_VARIANT* v, char** cstr)

    #ctypedef AMPL_Tuple AMPL_TUPLE
    ctypedef struct AMPL_TUPLE:
        pass

    void retainTuple(AMPL_TUPLE* t)

    void releaseTuple(AMPL_TUPLE* t)

    int AMPL_TupleCreate(AMPL_TUPLE** t, size_t size, AMPL_VARIANT** v)

    int AMPL_TupleCreateNumeric(AMPL_TUPLE** t, size_t size, double* v)

    int AMPL_TupleCreateString(AMPL_TUPLE** t, size_t size, const char* const* v)

    int AMPL_TupleCopy(AMPL_TUPLE** t, AMPL_TUPLE* copy)

    int AMPL_TupleFree(AMPL_TUPLE** t)

    int AMPL_TupleCompare(AMPL_TUPLE* t1, AMPL_TUPLE* t2)

    int AMPL_TupleGetSize(AMPL_TUPLE* t, size_t* size)

    int AMPL_TupleGetVariantPtr(AMPL_TUPLE* t, AMPL_VARIANT*** v)

    int AMPL_TupleGetVariant(AMPL_TUPLE* t, size_t index, AMPL_VARIANT** v)

    int AMPL_TupleToString(AMPL_TUPLE* t, char** cstr)

    #ctypedef AMPL_DataFrame AMPL_DATAFRAME
    ctypedef struct AMPL_DATAFRAME:
        pass

    AMPL_ERRORINFO* AMPL_DataFrameCreate(AMPL_DATAFRAME** dataframe, size_t numberOfIndexColumns, size_t numberOfDataColumns, const char* const* headers)

    AMPL_ERRORINFO* AMPL_DataFrameCreate2(AMPL_DATAFRAME** dataframe, size_t numberOfIndexColumns)

    AMPL_ERRORINFO* AMPL_DataFrameCopy(AMPL_DATAFRAME** dataframe, AMPL_DATAFRAME* copy)

    void AMPL_DataFrameFree(AMPL_DATAFRAME** dataframe)

    AMPL_ERRORINFO* AMPL_DataFrameGetHeaders(AMPL_DATAFRAME* dataframe, size_t* size, char*** headers)

    AMPL_ERRORINFO* AMPL_DataFrameEquals(AMPL_DATAFRAME* df1, AMPL_DATAFRAME* df2, int* equals)

    AMPL_ERRORINFO* AMPL_DataFrameToString(AMPL_DATAFRAME* dataframe, char** output)

    AMPL_ERRORINFO* AMPL_DataFrameReserve(AMPL_DATAFRAME* dataframe, size_t numRows)

    AMPL_ERRORINFO* AMPL_DataFrameAddRow(AMPL_DATAFRAME* dataframe, AMPL_TUPLE* value)

    AMPL_ERRORINFO* AMPL_DataFrameSetColumnArg(AMPL_DATAFRAME* dataframe, const char* header, AMPL_ARGS* column, size_t n)

    AMPL_ERRORINFO* AMPL_DataFrameSetColumnArgDouble(AMPL_DATAFRAME* dataframe, const char* header, const double* column, size_t n)

    AMPL_ERRORINFO* AMPL_DataFrameSetColumnArgString(AMPL_DATAFRAME* dataframe, const char* header, const char* const* column, size_t n)

    AMPL_ERRORINFO* AMPL_DataFrameSetValue(AMPL_DATAFRAME* dataframe, AMPL_TUPLE* rowIndex, const char* header, AMPL_VARIANT* value)

    AMPL_ERRORINFO* AMPL_DataFrameSetValueByIndex(AMPL_DATAFRAME* dataframe, size_t rowNumber, size_t colNumber, AMPL_VARIANT* value)

    AMPL_ERRORINFO* AMPL_DataFrameAddColumn(AMPL_DATAFRAME* dataframe, const char* header, AMPL_ARGS* values)

    AMPL_ERRORINFO* AMPL_DataFrameAddColumnDouble(AMPL_DATAFRAME* dataframe, const char* header, const double* values)

    AMPL_ERRORINFO* AMPL_DataFrameAddColumnString(AMPL_DATAFRAME* dataframe, const char* header, const char** values)

    AMPL_ERRORINFO* AMPL_DataFrameAddEmptyColumn(AMPL_DATAFRAME* dataframe, const char* header)

    AMPL_ERRORINFO* AMPL_DataFrameGetNumCols(AMPL_DATAFRAME* dataframe, size_t* num)

    AMPL_ERRORINFO* AMPL_DataFrameGetNumRows(AMPL_DATAFRAME* dataframe, size_t* num)

    AMPL_ERRORINFO* AMPL_DataFrameGetNumIndices(AMPL_DATAFRAME* dataframe, size_t* num)

    AMPL_ERRORINFO* AMPL_DataFrameSetArray(AMPL_DATAFRAME* dataframe, const double* values, size_t l0, AMPL_ARGS* indices0)

    AMPL_ERRORINFO* AMPL_DataFrameSetArrayString(AMPL_DATAFRAME* dataframe, const char* const* values, size_t l0, AMPL_ARGS* indices0)

    AMPL_ERRORINFO* AMPL_DataFrameSetMatrix(AMPL_DATAFRAME* dataframe, const double* values, size_t l0, AMPL_ARGS* indices0, size_t l1, AMPL_ARGS* indices1)

    AMPL_ERRORINFO* AMPL_DataFrameSetMatrixStringString(AMPL_DATAFRAME* dataframe, const double* values, size_t l0, const char* const* indices0, size_t l1, const char* const* indices1)

    AMPL_ERRORINFO* AMPL_DataFrameSetMatrixString(AMPL_DATAFRAME* dataframe, const char* const* values, size_t l0, AMPL_ARGS* indices0, size_t l1, AMPL_ARGS* indices1)

    AMPL_ERRORINFO* AMPL_DataFrameGetColumnIndex(AMPL_DATAFRAME* dataframe, const char* name, size_t* columnindex)

    AMPL_ERRORINFO* AMPL_DataFrameGetIndexingTuple(AMPL_DATAFRAME* dataframe, size_t rowindex, AMPL_TUPLE** index)

    AMPL_ERRORINFO* AMPL_DataFrameGetRowIndex(AMPL_DATAFRAME* dataframe, AMPL_TUPLE* index, size_t* rowindex)

    AMPL_ERRORINFO* AMPL_DataFrameElement(AMPL_DATAFRAME* dataframe, size_t rowindex, size_t colindex, AMPL_VARIANT** v)

    ctypedef struct AMPL_ENVIRONMENTVAR:
        char* name
        char* value

    int AMPL_EnvironmentVarGetName(AMPL_ENVIRONMENTVAR* envvar, char** name)

    int AMPL_EnvironmentVarGetValue(AMPL_ENVIRONMENTVAR* envvar, char** value)

    #ctypedef AMPL_Environment AMPL_ENVIRONMENT
    ctypedef struct AMPL_ENVIRONMENT:
        pass

    int AMPL_EnvironmentCreate(AMPL_ENVIRONMENT** env, const char* binaryDirectory, const char* binaryName)

    int AMPL_EnvironmentFree(AMPL_ENVIRONMENT** env)

    int AMPL_EnvironmentCopy(AMPL_ENVIRONMENT** copy, AMPL_ENVIRONMENT* src)

    int AMPL_EnvironmentAddEnvironmentVariable(AMPL_ENVIRONMENT* env, const char* name, const char* value)

    int AMPL_EnvironmentGetBinaryDirectory(AMPL_ENVIRONMENT* env, char** binaryDirectory)

    int AMPL_EnvironmentGetAMPLCommand(AMPL_ENVIRONMENT* env, char** amplCommand)

    int AMPL_EnvironmentSetBinaryDirectory(AMPL_ENVIRONMENT* env, const char* binaryDirectory)

    int AMPL_EnvironmentGetBinaryName(AMPL_ENVIRONMENT* env, char** binaryName)

    int AMPL_EnvironmentSetBinaryName(AMPL_ENVIRONMENT* env, const char* binaryName)

    int AMPL_EnvironmentToString(AMPL_ENVIRONMENT* env, char** str)

    int AMPL_EnvironmentGetSize(AMPL_ENVIRONMENT* env, size_t* size)

    int AMPL_EnvironmentGetEnvironmentVar(AMPL_ENVIRONMENT* env, AMPL_ENVIRONMENTVAR** envvar)

    int AMPL_EnvironmentFindEnvironmentVar(AMPL_ENVIRONMENT* env, const char* name, AMPL_ENVIRONMENTVAR** envvar)

    ctypedef enum AMPL_OUTPUTKIND:
        AMPL_OUTPUT_WAITING
        AMPL_OUTPUT_BREAK
        AMPL_OUTPUT_CD
        AMPL_OUTPUT_DISPLAY
        AMPL_OUTPUT_EXIT
        AMPL_OUTPUT_EXPAND
        AMPL_OUTPUT_LOAD
        AMPL_OUTPUT_OPTION
        AMPL_OUTPUT_PRINT
        AMPL_OUTPUT_PROMPT
        AMPL_OUTPUT_SOLUTION
        AMPL_OUTPUT_SOLVE
        AMPL_OUTPUT_SHOW
        AMPL_OUTPUT_XREF
        AMPL_OUTPUT_SHELL_OUTPUT
        AMPL_OUTPUT_SHELL_MESSAGE
        AMPL_OUTPUT_MISC
        AMPL_OUTPUT_WRITE_TABLE
        AMPL_OUTPUT_READ_TABLE
        AMPL_OUTPUT_READTABLE
        AMPL_OUTPUT_WRITETABLE
        AMPL_OUTPUT_BREAKPOINT
        AMPL_OUTPUT_CALL
        AMPL_OUTPUT_CHECK
        AMPL_OUTPUT_CLOSE
        AMPL_OUTPUT_COMMANDS
        AMPL_OUTPUT_CONTINUE
        AMPL_OUTPUT_DATA
        AMPL_OUTPUT_DELETECMD
        AMPL_OUTPUT_DROP
        AMPL_OUTPUT_DROP_OR_RESTORE_ALL
        AMPL_OUTPUT_ELSE
        AMPL_OUTPUT_ELSE_CHECK
        AMPL_OUTPUT_ENDIF
        AMPL_OUTPUT_ENVIRON
        AMPL_OUTPUT_FIX
        AMPL_OUTPUT_FOR
        AMPL_OUTPUT_IF
        AMPL_OUTPUT_LET
        AMPL_OUTPUT_LOOPEND
        AMPL_OUTPUT_OBJECTIVE
        AMPL_OUTPUT_OPTION_RESET
        AMPL_OUTPUT_PRINTF
        AMPL_OUTPUT_PROBLEM
        AMPL_OUTPUT_PURGE
        AMPL_OUTPUT_RBRACE
        AMPL_OUTPUT_READ
        AMPL_OUTPUT_RELOAD
        AMPL_OUTPUT_REMOVE
        AMPL_OUTPUT_REPEAT
        AMPL_OUTPUT_REPEAT_END
        AMPL_OUTPUT_RESET
        AMPL_OUTPUT_RESTORE
        AMPL_OUTPUT_RUN_ARGS
        AMPL_OUTPUT_SEMICOLON
        AMPL_OUTPUT_SSTEP
        AMPL_OUTPUT_THEN
        AMPL_OUTPUT_UNFIX
        AMPL_OUTPUT_UNLOAD
        AMPL_OUTPUT_UPDATE
        AMPL_OUTPUT_WRITE

    ctypedef void (*AMPL_OutputHandlerCb)(AMPL_OUTPUTKIND, const char*, void*) except *

    ctypedef void (*RunnablePtr)(void* runnable)

    int AMPL_StringFree(char** string)

    int AMPL_ErrorInfoFree(AMPL_ERRORINFO** error)

    void AMPL_AddToPath(const char* newPath)

    #ctypedef Ampl AMPL
    ctypedef struct AMPL:
        pass

    cdef struct ArrowSchema:
        pass

    #ctypedef ArrowSchema s_ArrowSchema
    
    cdef struct ArrowArray:
        pass

    #ctypedef ArrowArray s_ArrowArray

    cdef struct AMPL_DATAFRAMEARROW:
        pass

    AMPL_ERRORINFO *AMPL_DataFrameArrowCreate(AMPL_DATAFRAMEARROW **dataframe,
                                          ArrowSchema *schema, ArrowArray *array,
                                          int64_t nindices)

    void AMPL_DataFrameArrowFree(AMPL_DATAFRAMEARROW **dataframe)

    AMPL_ERRORINFO *AMPL_DataFrameArrowGetSchema(AMPL_DATAFRAMEARROW *dataframe, ArrowSchema **schema)

    AMPL_ERRORINFO *AMPL_DataFrameArrowGetArray(AMPL_DATAFRAMEARROW *dataframe, ArrowArray **array)

    AMPL_ERRORINFO *AMPL_DataFrameArrowGetNIndices(AMPL_DATAFRAMEARROW *dataframe, int64_t *nindices)

    AMPL_ERRORINFO* AMPL_Create(AMPL** ampl)

    AMPL_ERRORINFO* AMPL_CreateWithEnv(AMPL** ampl, AMPL_ENVIRONMENT* env)

    void AMPL_Free(AMPL** ampl)

    AMPL_ERRORINFO* AMPL_Eval(AMPL* ampl, const char* statement) except *

    AMPL_ERRORINFO* AMPL_EvalAsync(AMPL* ampl, const char* statement, RunnablePtr function, void* cb)

    AMPL_ERRORINFO* AMPL_SolveAsync(AMPL* ampl, RunnablePtr function, void* cb)

    AMPL_ERRORINFO* AMPL_ReadAsync(AMPL* ampl, const char* filename, RunnablePtr function, void* cb)

    AMPL_ERRORINFO* AMPL_ReadDataAsync(AMPL* ampl, const char* filename, RunnablePtr function, void* cb)

    AMPL_ERRORINFO* AMPL_Reset(AMPL* ampl)

    AMPL_ERRORINFO* AMPL_Close(AMPL* ampl)

    AMPL_ERRORINFO* AMPL_IsRunning(AMPL* ampl, bool* running)

    AMPL_ERRORINFO* AMPL_IsBusy(AMPL* ampl, bool* busy)

    AMPL_ERRORINFO* AMPL_Solve(AMPL* ampl, const char* problem, const char* solver)

    AMPL_ERRORINFO* AMPL_Interrupt(AMPL* ampl)

    AMPL_ERRORINFO* AMPL_Snapshot(AMPL* ampl, const char* fileName, bool model, bool data, bool options, char** output)

    AMPL_ERRORINFO* AMPL_ExportModel(AMPL* ampl, const char* fileName, char** output)

    AMPL_ERRORINFO* AMPL_ExportData(AMPL* ampl, const char* fileName, char** output)

    AMPL_ERRORINFO* AMPL_Cd(AMPL* ampl, char** output)

    AMPL_ERRORINFO* AMPL_Cd2(AMPL* ampl, const char* path, char** output)

    AMPL_ERRORINFO* AMPL_GetCurrentObjective(AMPL* ampl, char** currentObjective)

    AMPL_ERRORINFO* AMPL_SetOption(AMPL* ampl, const char* name, const char* value)

    AMPL_ERRORINFO* AMPL_GetOption(AMPL* ampl, const char* name, bool* exists, char** value)

    AMPL_ERRORINFO* AMPL_GetIntOption(AMPL* ampl, const char* name, bool* exists, int* value)

    AMPL_ERRORINFO* AMPL_GetDblOption(AMPL* ampl, const char* name, bool* exists, double* value)

    AMPL_ERRORINFO* AMPL_SetDblOption(AMPL* ampl, const char* name, double value)

    AMPL_ERRORINFO* AMPL_SetIntOption(AMPL* ampl, const char* name, int value)

    AMPL_ERRORINFO* AMPL_GetBoolOption(AMPL* ampl, const char* name, bool* exists, bool* value)

    AMPL_ERRORINFO* AMPL_SetBoolOption(AMPL* ampl, const char* name, bool value)

    AMPL_ERRORINFO* AMPL_Read(AMPL* ampl, const char* fileName)

    AMPL_ERRORINFO* AMPL_ReadData(AMPL* ampl, const char* fileName)

    AMPL_ERRORINFO* AMPL_GetData(AMPL* ampl, const char* const* displayStatements, size_t n, AMPL_DATAFRAME** output)

    AMPL_ERRORINFO *AMPL_GetDataArrow(AMPL *ampl, const char *const *displayStatements, size_t n, ArrowSchema **schema, ArrowArray **array)

    AMPL_ERRORINFO* AMPL_SetData(AMPL* ampl, AMPL_DATAFRAME* df, const char* setName)

    AMPL_ERRORINFO *AMPL_SetDataArrow(AMPL *ampl, AMPL_DATAFRAMEARROW *df, const char *setName)

    AMPL_ERRORINFO* AMPL_ToString(AMPL* ampl, char** output)

    AMPL_ERRORINFO* AMPL_ReadTable(AMPL* ampl, const char* tableName)

    AMPL_ERRORINFO* AMPL_WriteTable(AMPL* ampl, const char* tableName)

    AMPL_ERRORINFO* AMPL_Write(AMPL* ampl, const char* filename, const char* auxfiles)

    AMPL_ERRORINFO* AMPL_GetValue(AMPL* ampl, const char* scalarExpression, AMPL_VARIANT** v)

    AMPL_ERRORINFO* AMPL_GetValueString(AMPL* ampl, const char* scalarExpression, char** value)

    AMPL_ERRORINFO* AMPL_GetValueNumeric(AMPL* ampl, const char* scalarExpression, double* value)

    AMPL_ERRORINFO* AMPL_GetOutput(AMPL* ampl, const char* amplstatement, char** output)

    AMPL_ERRORINFO* AMPL_CallVisualisationCommandOnNames(AMPL* ampl, const char* command, const char* const* args, size_t nargs)

    AMPL_ERRORINFO* AMPL_SetOutputHandler(AMPL* ampl, AMPL_OutputHandlerCb callback, void* usrdata)

    AMPL_ERRORINFO* AMPL_SetErrorHandler(AMPL* ampl, ErrorHandlerCbPtr callback, void* usrdata)

    AMPL_ERRORINFO *AMPL_GetOutputHandler(AMPL *ampl, void **usrdata);

    AMPL_ERRORINFO *AMPL_GetErrorHandler(AMPL *ampl, void **usrdata);

    AMPL_ERRORINFO* AMPL_GetVariables(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetConstraints(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetParameters(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetObjectives(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetSets(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetTables(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_GetProblems(AMPL* ampl, size_t* size, char*** names)

    AMPL_ERRORINFO* AMPL_DataFrameCreate3(AMPL_DATAFRAME** dataframe, AMPL* ampl, const char* const* args, size_t nargs)

    ctypedef enum AMPL_ENTITYTYPE:
        AMPL_VARIABLE
        AMPL_CONSTRAINT
        AMPL_OBJECTIVE
        AMPL_PARAMETER
        AMPL_SET
        AMPL_TABLE
        AMPL_PROBLEM
        AMPL_UNDEFINED

    AMPL_ERRORINFO* AMPL_EntityGetIndexarity(AMPL* ampl, const char* name, size_t* arity)

    AMPL_ERRORINFO* AMPL_EntityGetXref(AMPL* ampl, const char* name, char*** xref, size_t* size)

    AMPL_ERRORINFO* AMPL_EntityGetNumInstances(AMPL* ampl, const char* name, size_t* size)

    AMPL_ERRORINFO* AMPL_EntityGetTuples(AMPL* ampl, const char* name, AMPL_TUPLE*** tuples, size_t* size)

    AMPL_ERRORINFO* AMPL_EntityGetIndexingSets(AMPL* ampl, const char* name, char*** indexingsets, size_t* size)

    AMPL_ERRORINFO* AMPL_EntityGetType(AMPL* ampl, const char* name, AMPL_ENTITYTYPE* type)

    AMPL_ERRORINFO* AMPL_EntityGetTypeString(AMPL* ampl, const char* name, const char** typestr)

    AMPL_ERRORINFO* AMPL_EntityGetDeclaration(AMPL* ampl, const char* name, char** declaration)

    AMPL_ERRORINFO* AMPL_EntityDrop(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_EntityRestore(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_EntityGetValues(AMPL* ampl, const char* name, const char* const* suffixes, size_t n, AMPL_DATAFRAME** output)

    AMPL_ERRORINFO* AMPL_EntitySetValues(AMPL* ampl, const char* name, AMPL_DATAFRAME* data)

    AMPL_ERRORINFO* AMPL_EntitySetValuesArrow(AMPL *ampl, const char *name, AMPL_DATAFRAMEARROW *dataframe)

    AMPL_ERRORINFO* AMPL_ParameterSetValue(AMPL* ampl, const char* scalarExpression, AMPL_VARIANT* v)

    AMPL_ERRORINFO* AMPL_ParameterSetNumeric(AMPL* ampl, const char* scalarExpression, double value)

    AMPL_ERRORINFO* AMPL_ParameterSetString(AMPL* ampl, const char* scalarExpression, const char* value)

    AMPL_ERRORINFO* AMPL_ParameterIsSymbolic(AMPL* ampl, const char* name, bool* isSymbolic)

    AMPL_ERRORINFO* AMPL_ParameterHasDefault(AMPL* ampl, const char* name, bool* hasDefault)

    AMPL_ERRORINFO* AMPL_ParameterSetArgsDoubleValues(AMPL* ampl, const char* name, size_t size, const double* args)

    AMPL_ERRORINFO* AMPL_ParameterSetArgsStringValues(AMPL* ampl, const char* name, size_t size, const char* const* args)

    AMPL_ERRORINFO* AMPL_ParameterSetArgsValues(AMPL* ampl, const char* name, size_t size, AMPL_ARGS* args)

    AMPL_ERRORINFO* AMPL_ParameterSetValuesMatrix(AMPL* ampl, const char* name, size_t nrows, AMPL_ARGS* row_indices, size_t ncols, AMPL_ARGS* col_indices, const double* data, bool transpose)

    AMPL_ERRORINFO* AMPL_ParameterSetSomeValues(AMPL* ampl, const char* name, size_t size, AMPL_TUPLE** index, AMPL_VARIANT** v)

    AMPL_ERRORINFO* AMPL_ParameterSetSomeArgsValues(AMPL* ampl, const char* name, size_t size, AMPL_TUPLE** index, AMPL_ARGS* args)

    AMPL_ERRORINFO *AMPL_ParameterSetSomeStringValues(AMPL *ampl, const char *name, size_t size, AMPL_TUPLE **index, char **str_values)

    AMPL_ERRORINFO *AMPL_ParameterSetSomeDoubleValues(AMPL *ampl, const char *name, size_t size, AMPL_TUPLE **index, double *dbl_values)

    AMPL_ERRORINFO* AMPL_VariableGetValue(AMPL* ampl, const char* name, double* value)

    AMPL_ERRORINFO* AMPL_VariableFix(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_VariableFixWithValue(AMPL* ampl, const char* name, double value)

    AMPL_ERRORINFO* AMPL_VariableUnfix(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_VariableSetValue(AMPL* ampl, const char* name, double value)

    AMPL_ERRORINFO* AMPL_VariableGetIntegrality(AMPL* ampl, const char* name, int* integrality)

    AMPL_ERRORINFO* AMPL_ConstraintIsLogical(AMPL* ampl, const char* name, bool* isLogical)

    AMPL_ERRORINFO* AMPL_ConstraintSetDual(AMPL* ampl, const char* name, double dual)

    AMPL_ERRORINFO* AMPL_ObjectiveSense(AMPL* ampl, const char* name, int* sense)

    AMPL_ERRORINFO* AMPL_SetGetArity(AMPL* ampl, const char* name, size_t* arity)

    AMPL_ERRORINFO* AMPL_TableRead(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_TableWrite(AMPL* ampl, const char* name)

    AMPL_ERRORINFO* AMPL_InstanceGetDoubleSuffix(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_NUMERICSUFFIX suffix, double* value)

    AMPL_ERRORINFO* AMPL_InstanceGetIntSuffix(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_NUMERICSUFFIX suffix, int* value)

    AMPL_ERRORINFO* AMPL_InstanceGetStringSuffix(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_STRINGSUFFIX suffix, char** value)

    AMPL_ERRORINFO* AMPL_InstanceGetName(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, char** name)

    AMPL_ERRORINFO* AMPL_InstanceToString(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, char** str)

    AMPL_ERRORINFO* AMPL_InstanceDrop(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)

    AMPL_ERRORINFO* AMPL_InstanceRestore(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)

    AMPL_ERRORINFO* AMPL_VariableInstanceFix(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)

    AMPL_ERRORINFO* AMPL_VariableInstanceFixToValue(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, double value)

    AMPL_ERRORINFO* AMPL_VariableInstanceUnfix(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)

    AMPL_ERRORINFO* AMPL_VariableInstanceSetValue(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, double value)

    AMPL_ERRORINFO* AMPL_VariableInstanceToString(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, char** str)

    AMPL_ERRORINFO* AMPL_ConstraintInstanceSetDual(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, double dual)

    AMPL_ERRORINFO* AMPL_SetInstanceGetSize(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, size_t* size)

    AMPL_ERRORINFO* AMPL_SetInstanceContains(AMPL* ampl, const char* entityname, AMPL_TUPLE* index, AMPL_TUPLE* tuple, bool* contains)

    AMPL_ERRORINFO* AMPL_SetInstanceGetValues(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_TUPLE*** tuples, size_t* size)

    AMPL_ERRORINFO* AMPL_SetInstanceGetValuesDataframe(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_DATAFRAME** dataframe)

    AMPL_ERRORINFO *AMPL_SetInstanceGetValuesArrow(AMPL *ampl, const char *entityname, AMPL_TUPLE *tuple,  AMPL_DATAFRAMEARROW **dataframe_arrow)

    AMPL_ERRORINFO* AMPL_SetInstanceSetValues(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_ARGS* args, size_t size)

    AMPL_ERRORINFO* AMPL_SetInstanceSetValuesTuples(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_TUPLE** tuples, size_t size)

    AMPL_ERRORINFO* AMPL_SetInstanceSetValuesDataframe(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, AMPL_DATAFRAME* data)

    AMPL_ERRORINFO *AMPL_SetInstanceSetValuesArrow(AMPL *ampl, const char *entityname, AMPL_TUPLE *tuple, AMPL_DATAFRAMEARROW *dataframe);

    AMPL_ERRORINFO* AMPL_SetInstanceToString(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple, char** str)

    AMPL_ERRORINFO* AMPL_ParameterInstanceSetValue(AMPL* ampl, const char* name, AMPL_TUPLE* index, AMPL_VARIANT* v)

    AMPL_ERRORINFO* AMPL_ParameterInstanceSetNumericValue(AMPL* ampl, const char* name, AMPL_TUPLE* index, double value)

    AMPL_ERRORINFO* AMPL_ParameterInstanceSetStringValue(AMPL* ampl, const char* name, AMPL_TUPLE* index, const char* value)

    AMPL_ERRORINFO* AMPL_TableInstanceRead(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)

    AMPL_ERRORINFO* AMPL_TableInstanceWrite(AMPL* ampl, const char* entityname, AMPL_TUPLE* tuple)



# -*- coding: utf-8 -*-


cdef class Objective(Entity):
    """
    Bases: :class:`~amplpy.Entity`.

    Represents an AMPL objective. Note that, in case of a scalar objective, all
    the properties (corresponding to AMPL suffixes) of the objective instance
    can be accessed through methods like :func:`~amplpy.Objective.value`.
    The methods have the same name of the corresponding AMPL suffixes.
    See http://www.ampl.com/NEW/suffbuiltin.html for a list of the available
    suffixes.

    All these methods throw a TypeError if called for a non scalar
    objective and an RuntimeError if called on an entity which has been deleted
    in the underlying intepreter.

    To gain access to all the values in an entity (for all instances and all
    suffixes for that entities), see
    :func:`~amplpy.Entity.get_values`
    and the :class:`~amplpy.DataFrame` class.
    """
    @staticmethod
    cdef create(AMPL ampl, char* name, campl.AMPL_TUPLE* index, parent):
        entity = Objective()
        entity._ampl = ampl
        Py_INCREF(entity._ampl)
        entity._name = name
        entity._index = index
        entity.wrap_function = campl.AMPL_OBJECTIVE
        entity._entity = parent
        if entity._entity is not None:
            Py_INCREF(entity._entity)
        return entity

    def value(self):
        """
        Get the value of the objective.
        """
        cdef double value
        PY_AMPL_CALL(campl.AMPL_InstanceGetDoubleSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_VALUE, &value))
        return value

    def astatus(self):
        """
        Return the AMPL status.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_ASTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def sstatus(self):
        """
        Return the solver status.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_SSTATUS, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def exitcode(self):
        """
        Exit code returned by solver after most recent solve with this
        objective.
        """
        cdef int value
        PY_AMPL_CALL(campl.AMPL_InstanceGetIntSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_NUMERICSUFFIX.AMPL_EXITCODE, &value))
        return value

    def message(self):
        """
        Result message returned by solver after most recent solve with this
        objective.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_MESSAGE, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def result(self):
        """
        Result string returned by solver after most recent solve with this
        objective.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_RESULT, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        return value

    def drop(self):
        """
        Drop this objective instance.
        """
        PY_AMPL_CALL(campl.AMPL_EntityDrop(self._ampl._c_ampl, self._name))

    def restore(self):
        """
        Restore this objective (if it had been dropped, no effect otherwise).
        """
        PY_AMPL_CALL(campl.AMPL_EntityRestore(self._ampl._c_ampl, self._name))

    def minimization(self):
        """
        Get the sense of this objective. Returns True if minimize,
        False if maximize.
        """
        cdef campl.AMPL_ERRORINFO* errorinfo
        cdef char* value_c
        errorinfo = campl.AMPL_InstanceGetStringSuffix(self._ampl._c_ampl, self._name, self._index, campl.AMPL_STRINGSUFFIX.AMPL_SENSE, &value_c)
        if errorinfo:
            PY_AMPL_CALL(errorinfo)
        value = str(value_c.decode('utf-8'))
        campl.AMPL_StringFree(&value_c)

        if value == 'minimize':
            return True
        else:
            return False



# -*- coding: utf-8 -*-


cdef class AMPLException(Exception):
    """
    Represent an exception generated by AMPL.
    """

    def __init__(self, source_name, line_number, offset, message):
        super(AMPLException, self).__init__()
        self.source_name = source_name
        self.line_number = line_number
        self.offset = offset
        self.message = message

    def get_source_name(self):
        """
        Get the name of the file where the error was detected.
        """
        return self.source_name

    def get_line_number(self):
        """
        Get the row where the error is located.
        """
        return self.line_number

    def get_offset(self):
        """
        Get the offset where the error is located.
        """
        return self.offset

    def get_message(self):
        """
        Get the error message.
        """
        return self.message

    def __str__(self):
        if self.offset == -1:
            return self.message
        elif self.source_name == "-":
            return "line " + str(self.line_number) + " offset " + str(self.offset) + "\n" + self.message
        else:
            return self.source_name + "\nline " + str(self.line_number) + " offset " + str(self.offset) + "\n" + self.message

    # Aliases
    getLineNumber = get_line_number
    getMessage = get_message
    getOffset = get_offset
    getSourceName = get_source_name


class PresolveException(RuntimeError):
    pass


class InfeasibilityException(RuntimeError):
    pass



from cpython.dict cimport PyDict_Check, PyDict_Keys, PyDict_Values
from cpython.unicode cimport PyUnicode_Check
from cpython.unicode cimport PyUnicode_AsUTF8
from cpython.list cimport PyList_GetItem
from cpython.long cimport PyLong_Check, PyLong_AsLong
from cpython.float cimport PyFloat_AsDouble
from cpython.object cimport PyObject
from libcpp cimport bool

cdef void PY_AMPL_CALL(campl.AMPL_ERRORINFO* errorinfo) except *:
    cdef campl.AMPL_ERRORCODE rc
    cdef char* message
    cdef char* source
    if errorinfo:
        rc = campl.AMPL_ErrorInfoGetError(errorinfo)
        if rc == campl.AMPL_INFEASIBILITY_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise InfeasibilityException("InfeasibilityException: " + message.decode('utf-8'))
        elif rc == campl.AMPL_PRESOLVE_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise PresolveException("PresolveException: " + message.decode('utf-8'))
        elif rc == campl.AMPL_LICENSE_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise SystemError(message.decode('utf-8'))
        elif rc == campl.AMPL_FILE_IO_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise IOError(message.decode('utf-8'))
        elif rc == campl.AMPL_UNSUPPORTED_OPERATION_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise TypeError(message.decode('utf-8'))
        elif rc == campl.AMPL_INVALID_SUBSCRIPT_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            source = campl.AMPL_ErrorInfoGetSource(errorinfo)
            raise AMPLException(source.decode('utf-8'), campl.AMPL_ErrorInfoGetLine(errorinfo), campl.AMPL_ErrorInfoGetOffset(errorinfo),
                                          message.decode('utf-8'))
        elif rc == campl.AMPL_SYNTAX_ERROR_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            source = campl.AMPL_ErrorInfoGetSource(errorinfo)
            raise AMPLException(source.decode('utf-8'), campl.AMPL_ErrorInfoGetLine(errorinfo), campl.AMPL_ErrorInfoGetOffset(errorinfo),
                                     message.decode('utf-8'))
        elif rc == campl.AMPL_NO_DATA_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise RuntimeError(message.decode('utf-8'))
        elif rc == campl.AMPL_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise RuntimeError(message.decode('utf-8'))
        elif rc == campl.AMPL_RUNTIME_ERROR:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise RuntimeError(message.decode('utf-8'))
        elif rc == campl.AMPL_LOGIC_ERROR:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise TypeError(message.decode('utf-8'))
        elif rc == campl.AMPL_OUT_OF_RANGE:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise KeyError(message.decode('utf-8'))
        elif rc == campl.AMPL_INVALID_ARGUMENT:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise ValueError(message.decode('utf-8'))
        elif rc == campl.AMPL_STD_EXCEPTION:
            message = campl.AMPL_ErrorInfoGetMessage(errorinfo)
            raise RuntimeError(message.decode('utf-8'))
        else:
            raise RuntimeError("Unknown exception")

cdef campl.AMPL_ERRORINFO* setValues(campl.AMPL* ampl, char* name, campl.AMPL_TUPLE* index, values, size_t size):
    cdef campl.AMPL_ERRORINFO* errorinfo
    cdef campl.AMPL_TUPLE** values_c
    cdef size_t i
    values_c = <campl.AMPL_TUPLE**>malloc(size * sizeof(campl.AMPL_TUPLE*))

    for i in range(size):
        values_c[i] = to_c_tuple(values[i])
    
    errorinfo = campl.AMPL_SetInstanceSetValuesTuples(ampl, name, index, values_c, size)

    for i in range(size):
        campl.AMPL_TupleFree(&values_c[i])
    free(values_c)

    return errorinfo

cdef to_py_variant(campl.AMPL_VARIANT* variant):
    cdef campl.AMPL_TYPE type
    cdef char* value_c
    cdef double real_c
    campl.AMPL_VariantGetType(variant, &type)
    if type == campl.AMPL_STRING:
        campl.AMPL_VariantGetStringValue(variant, &value_c)
        return value_c.decode('utf-8')
    elif type == campl.AMPL_NUMERIC:
        campl.AMPL_VariantGetNumericValue(variant, &real_c)
        if real_c % 1 == 0:
            return int(real_c)
        else:
            return real_c
    else:
        return None

cdef to_py_tuple(campl.AMPL_TUPLE* tuple_c):
    cdef size_t size
    cdef campl.AMPL_VARIANT** variants
    cdef list pylist = []
    campl.AMPL_TupleGetSize(tuple_c, &size)
    campl.AMPL_TupleGetVariantPtr(tuple_c, &variants)
    for i in range(size):
        pylist.append(to_py_variant(variants[i]))
    return tuple(pylist)

cdef campl.AMPL_TUPLE* to_c_tuple(py_tuple):
    cdef size_t i
    cdef campl.AMPL_TUPLE* tuple_c

    if not isinstance(py_tuple, (tuple, list)):
        py_tuple = [py_tuple]

    cdef size_t size = len(py_tuple)
    cdef campl.AMPL_VARIANT** variants = <campl.AMPL_VARIANT**> malloc(size * sizeof(campl.AMPL_VARIANT*))

    for i in range(size):
        variants[i] = to_c_variant(py_tuple[i])

    campl.AMPL_TupleCreate(&tuple_c, size, variants)

    for i in range(size):
        campl.AMPL_VariantFree(&variants[i])

    free(variants)
    return tuple_c

cdef campl.AMPL_VARIANT* to_c_variant(value)  except *:
    cdef campl.AMPL_VARIANT* variant
    if isinstance(value, str):
        campl.AMPL_VariantCreateString(&variant, value.encode('utf-8'))
    elif isinstance(value, Real):
        campl.AMPL_VariantCreateNumeric(&variant, value)
    elif value is None:
        campl.AMPL_VariantCreateEmpty(&variant)
    else:
        raise ValueError(f"unsupported type {type(value)}")
    return variant

cdef create_entity(campl.AMPL_ENTITYTYPE entity_class, AMPL ampl, char* name, campl.AMPL_TUPLE* index, parent):
    if entity_class == campl.AMPL_VARIABLE:
        return Variable.create(ampl, name, index, parent)
    elif entity_class == campl.AMPL_CONSTRAINT:
        return Constraint.create(ampl, name, index, parent)
    elif entity_class == campl.AMPL_OBJECTIVE:
        return Objective.create(ampl, name, index, parent)
    elif entity_class == campl.AMPL_SET:
        return Set.create(ampl, name, index, parent)
    elif entity_class == campl.AMPL_PARAMETER:
        return Parameter.create(ampl, name, index, parent)
    else:
        return Entity.create(ampl, name, index, parent)

cdef campl.AMPL_ERRORINFO* setValuesParamNum(campl.AMPL* ampl, char* name, values):
    cdef campl.AMPL_ERRORINFO* errorinfo
    cdef size_t size = len(values)
    cdef double* values_c = <double*> malloc(size * sizeof(double))
    for i in range(size):
        values_c[i] = values[i]
    errorinfo = campl.AMPL_ParameterSetArgsDoubleValues(ampl, name, size, values_c)
    free(values_c)

    return errorinfo

cdef campl.AMPL_ERRORINFO* setValuesParamStr(campl.AMPL* ampl, char* name, values):
    cdef campl.AMPL_ERRORINFO* errorinfo
    cdef size_t size = len(values)
    cdef char** values_c = <char**> malloc(size * sizeof(char*))
    for i in range(size):
        values_c[i] = strdup(values[i].encode('utf-8'))
    
    errorinfo = campl.AMPL_ParameterSetArgsStringValues(ampl, name, size, values_c)

    for i in range(size):
        free(values_c[i])
    free(values_c)

    return errorinfo

cdef campl.AMPL_ERRORINFO* setValuesPyDict(campl.AMPL* ampl, char* name, dict dicts) except *:
    cdef campl.AMPL_ERRORINFO* errorinfo
    cdef size_t i
    cdef campl.AMPL_TUPLE** indices_c
    cdef char** values_str_c
    cdef double* values_num_c
    cdef PyObject* item

    if not PyDict_Check(dicts):
        raise ValueError("Expected a dictionary")
        
    cdef object d_keys = PyDict_Keys(dicts)
    cdef object d_values = PyDict_Values(dicts)
    cdef size_t size = len(dicts)

    cdef bool has_numbers = False
    cdef bool has_strings = False

    for i in range(size):
        item = PyList_GetItem(d_values, i)
        if item == NULL:
            raise ValueError("Failed to access value")

        if PyUnicode_Check(<object>item):
            has_strings = True
        else:
            has_numbers = True

        if has_numbers and has_strings:
            raise ValueError("All values must be either numbers or strings")
        
    if has_strings and not has_numbers:
        indices_c = <campl.AMPL_TUPLE**> malloc(size * sizeof(campl.AMPL_TUPLE*))
        values_str_c = <char **> malloc(size * sizeof(char*))
        for i in range(size):
            indices_c[i] = to_c_tuple(<object>PyList_GetItem(d_keys, i))
            values_str_c[i] = PyUnicode_AsUTF8(<object>PyList_GetItem(d_values, i))
        errorinfo = campl.AMPL_ParameterSetSomeStringValues(ampl, name, size, indices_c, values_str_c)
        for i in range(size):
            campl.AMPL_TupleFree(&indices_c[i])
        free(indices_c)
        free(values_str_c)
    elif has_numbers and not has_strings:
        indices_c = <campl.AMPL_TUPLE**> malloc(size * sizeof(campl.AMPL_TUPLE*))
        values_num_c = <double *> malloc(size * sizeof(double))
        for i in range(size):
            indices_c[i] = to_c_tuple(<object>PyList_GetItem(d_keys, i))
            item = PyList_GetItem(d_values, i)
            if PyLong_Check(<object>item):
                values_num_c[i] = PyLong_AsLong(<object>item)
            else:
                values_num_c[i] = PyFloat_AsDouble(<object>item)
        errorinfo = campl.AMPL_ParameterSetSomeDoubleValues(ampl, name, size, indices_c, values_num_c)
        for i in range(size):
            campl.AMPL_TupleFree(&indices_c[i])
        free(indices_c)
        free(values_num_c)
    else:
        raise ValueError("Dictionary must contain either all strings or all numbers")
    return errorinfo

cdef void raiseKeyError(campl.AMPL_ENTITYTYPE entity_class, str name) except *:
    if entity_class == campl.AMPL_VARIABLE:
        raise KeyError(f"A variable called {name} cannot be found.")
    elif entity_class == campl.AMPL_CONSTRAINT:
        raise KeyError(f"A constraint called {name} cannot be found.")
    elif entity_class == campl.AMPL_OBJECTIVE:
        raise KeyError(f"An objective called {name} cannot be found.")
    elif entity_class == campl.AMPL_SET:
        raise KeyError(f"A set called {name} cannot be found.")
    elif entity_class == campl.AMPL_PARAMETER:
        raise KeyError(f"A parameter called {name} cannot be found.")
    else: 
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
import string
import random
import os
from amplpy import DataFrameArrow
from . import TestBase

try:
    import pyarrow as pa
except ImportError:
    pa = None
try:
    import pandas as pd
except ImportError:
    pd = None
try:
    import nanoarrow as na
except ImportError:
    na = None

class TestDataFrameArrow(TestBase.TestBase):
    """Test DataFrame."""

    def test_pandas_arrow(self):
        ampl = self.ampl
        if pd is None:
            self.skipTest("pandas not available")
        if pa is None:
            self.skipTest("pyarrow not available")
        if na is None:
            self.skipTest("nanoarrow not available")
        df = pd.DataFrame({"a": [1, 2], "b": [3.5, 4]}, index=["x", "y"])
        ampl.eval(
            """
            set S;
            param a{S};
            param b{S};
        """
        )
        ampl.set_data_arrow(df, "S")
        self.assertEqual(list(ampl.set["S"].members()), ["x", "y"])
        self.assertEqual(ampl.param["a"].num_instances(), 2)
        self.assertEqual(ampl.param["a"]["x"], 1)
        self.assertEqual(ampl.param["a"]["y"], 2)
        self.assertEqual(ampl.param["b"].num_instances(), 2)
        self.assertEqual(ampl.param["b"]["x"], 3.5)
        self.assertEqual(ampl.param["b"]["y"], 4)

    def test_pandas_arrow2(self):
        ampl = self.ampl
        #if pd is None:
        #    self.skipTest("pandas not available")
        #if pa is None:
        #    self.skipTest("pyarrow not available")
        #if na is None:
        #    self.skipTest("nanoarrow not available")
        df = pd.DataFrame({"a": [1, 2], "b": [3.5, 4]}, index=["x", "y"])
        ampl.eval(
            """
            set S;
            param a{S};
            param b{S};
        """
        )
        ampl.set_data_arrow(df, "S")
        self.assertEqual(list(ampl.set["S"].members()), ["x", "y"])
        self.assertEqual(ampl.param["a"]["x"], 1)
        self.assertEqual(ampl.param["a"]["y"], 2)
        self.assertEqual(ampl.param["b"]["x"], 3.5)
        self.assertEqual(ampl.param["b"]["y"], 4)

        ampl.eval("reset data;")
        df2 = pd.DataFrame(
            {
                "a": [10, 20, 30],
            },
            index=["x", "y", "z"],
        )
        ampl.set_data_arrow(df2, "S")
        self.assertEqual(list(ampl.set["S"].members()), ["x", "y", "z"])
        self.assertEqual(ampl.param["a"]["x"], 10)
        self.assertEqual(ampl.param["a"]["y"], 20)
        self.assertEqual(ampl.param["a"]["z"], 30)
        
        
        ampl.eval("reset data;")
        df3 = pd.DataFrame({}, index=["xx", "yy"])
        ampl.set_data_arrow(df3, "S")
        self.assertEqual(list(ampl.set["S"].members()), ["xx", "yy"])

        csv_file = os.path.join(os.path.dirname(__file__), "data.csv")
        p_df = pd.read_csv(csv_file, sep=";", index_col=0)
        #df = DataFrame.from_pandas(p_df)
        #self.assertTrue(isinstance(df.to_dict(), dict))
        #self.assertEqual(set(df.to_dict().keys()), set([1.0, 2.0, 3.0]))
        #self.assertEqual(set(df.to_list()[0]), set([1.0, 0.01]))
        #self.assertEqual(set(df.to_list()[1]), set([2.0, 0.02]))
        #self.assertEqual(set(df.to_list()[2]), set([3.0, 0.03]))

    def test_setindex(self):
        ampl = self.ampl
        ampl.eval(
            """
            set NUTR;
            param n_min {NUTR} >= 0;
            param n_max {i in NUTR} >= n_min[i];
            """
            )
        nutr_df = pd.DataFrame(
            [
                ("A", 700, 20000),
                ("C", 500, 3000),
                ("B1", 100, 777),
                ("B2", 60, 500)
            ],
            columns=["NUTR", "n_min", "n_max"],
        ).set_index("NUTR")
        ampl.set_data_arrow(nutr_df, "NUTR")
        self.assertEqual(list(ampl.set["NUTR"].members()), ["A", "C", "B1", "B2"])
        self.assertEqual(ampl.param["n_min"]["A"], 700)
        self.assertEqual(ampl.param["n_min"]["C"], 500)
        self.assertEqual(ampl.param["n_min"]["B1"], 100)
        self.assertEqual(ampl.param["n_min"]["B2"], 60)
        self.assertEqual(ampl.param["n_max"]["A"], 20000)
        self.assertEqual(ampl.param["n_max"]["C"], 3000)
        self.assertEqual(ampl.param["n_max"]["B1"], 777)
        self.assertEqual(ampl.param["n_max"]["B2"], 500)

